#!/usr/env/bin python3
'''
Usage
=====
To merge larpix raw-formatted hdf5 files together::

    rawhdf5_tool.py --merge -i <files to merge, in order> -o <destination filename>

To split a single larpix raw-formatted hdf5 file into many::

    rawhdf5_tool.py --split -i <file to split> -o <destination directory> --max_length=<dataset length to split>

> Note:
>
> When merging files, there is the possibility of losing file metadata that is
> necessary to parse the file. The merge behavior will only keep the metadata of each
> dataset coming from the first file. It is highly recommended that you only
> `merge` files that were `split` using this utility, or files where the metadata
> guaranteed to be the same (i.e. many files of the same simulation run).

'''
import h5py
import warnings
import os
try:
    from tqdm import tqdm
    _has_tqdm = True
except Exception as e:
    warnings.warn(str(e), RuntimeWarning)
    _has_tqdm = False

_default_max_length = -1
_default_block_size = 102400

def merge_files(input_filenames, output_filename, block_size):
    with h5py.File(output_filename, 'w', libver='latest') as fo:
        curr_idx = 0
        prev_idx = 0
        for i,input_filename in enumerate(input_filenames):
            print(input_filename, '{}/{}'.format(i+1, len(input_filenames)))
            with h5py.File(input_filename, 'r', libver='latest', swmr=True) as fi:
                if i == 0:
                    # create datasets and groups
                    fo.create_group('meta')
                    fo.create_dataset('msgs', shape=(0,), maxshape=(None,), compression='gzip', dtype=fi['msgs'].dtype)
                    fo.create_dataset('msg_headers', shape=(0,), maxshape=(None,), compression='gzip', dtype=fi['msg_headers'].dtype)

                    # copy meta data
                    for attr,value in fi['meta'].attrs.items():
                        fo['meta'].attrs[attr] = value

                # resize datasets
                fo['msgs'].resize((len(fo['msgs']) + len(fi['msgs']),))
                fo['msg_headers'].resize((len(fo['msg_headers']) + len(fi['msg_headers']),))

                # copy data in chunks
                for start in tqdm(range(0, len(fi['msgs']), block_size)) if _has_tqdm else range(0, len(fi['msgs']), block_size):
                    end = min(start+block_size, len(fi['msgs']))
                    prev_idx = curr_idx
                    curr_idx += end - start
                    fo['msgs'][prev_idx:curr_idx] = [fi['msgs'][i] for i in range(start,end)] # loop needed because h5py was giving an error
                    fo['msg_headers'][prev_idx:curr_idx] = fi['msg_headers'][start:end]
    return

def split_file(input_filename, output_directory, max_length, block_size):
    with h5py.File(input_filename, 'r', libver='latest', swmr=True) as fi:
        output_filename_fmt = os.path.join(output_directory, os.path.basename(input_filename)[:-2]) + '{}.h5'

        i = 0
        curr_idx, prev_idx = 0, 0
        fo = None
        output_filename = output_filename_fmt.format(i)
        try:
            for start in tqdm(range(0, len(fi['msgs']), block_size)) if _has_tqdm else range(0, len(fi['msgs']), block_size):
                end = min(start+block_size, len(fi['msgs']))

                if fo is None or (os.stat(output_filename).st_size >= max_length and max_length > 0):
                    # open next file
                    if fo is not None:
                        fo.close()
                    output_filename = output_filename_fmt.format(i)
                    fo = h5py.File(output_filename, 'a', libver='latest')
                    i += 1
                    curr_idx, prev_idx = 0, 0

                    # create datasets and groups
                    fo.create_group('meta')
                    fo.create_dataset('msgs', shape=(0,), maxshape=(None,), compression='gzip', dtype=fi['msgs'].dtype)
                    fo.create_dataset('msg_headers', shape=(0,), maxshape=(None,), compression='gzip', dtype=fi['msg_headers'].dtype)

                    # copy meta data
                    for attr,value in fi['meta'].attrs.items():
                        fo['meta'].attrs[attr] = value

                    fo.swmr_mode = True

                # resize datasets
                prev_idx = curr_idx
                curr_idx += end - start
                fo['msgs'].resize((curr_idx,))
                fo['msg_headers'].resize((curr_idx,))

                # copy data
                fo['msgs'][prev_idx:curr_idx] = [fi['msgs'][i] for i in range(start,end)] # loop needed because h5py was giving an error
                fo['msg_headers'][prev_idx:curr_idx] = fi['msg_headers'][start:end]

                fo['msgs'].flush()
                fo['msg_headers'].flush()
        except:
            raise
        finally:
            if fo is not None:
                fo.close()
    return

def main(input_filenames, output_filename, max_length=_default_max_length, block_size=_default_block_size, **kwargs):
    if kwargs.get('merge', False):
        merge_files(
            input_filenames, output_filename,
            block_size=block_size
            )
    elif kwargs.get('split', False):
        split_file(
            input_filenames[0], output_filename,
            max_length=max_length, block_size=block_size
            )
    else:
        print('No action specified, exiting.')

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-i', nargs='+', required=True, help='Input file(s)')
    parser.add_argument('-o', required=True, type=str, help='Output file or directory')
    parser.add_argument('--max_length', type=int, default=_default_max_length, required=False, help='Max dataset length (split only, default=%(default)s bytes)')
    parser.add_argument('--block_size', type=int, default=_default_block_size, required=False, help='Block size used for reads (default=%(default)s)')
    parser.add_argument('--merge', action='store_true', help='Flag to merge files')
    parser.add_argument('--split', action='store_true', help='Flag to split files')
    args = parser.parse_args()
    main(
        input_filenames=args.i,
        output_filename=args.o,
        **vars(args)
        )



from bitarray import bitarray
import os
import errno
import functools
from collections import OrderedDict

from .. import bitarrayhelper as bah
from .. import configs
from . import Configuration_v2, _Smart_List
from . import configuration_v2 as conf_v2
from . import configuration_v2_base as v2_base

class Configuration_Lightpix_v1(Configuration_v2):
    '''
    Represents the desired configuration state of a LightPix v1 chip.

    This is a small extension of the LArPix v2 configuration register space to
    include the additional registers associated with LightPix, see the v2
    configuration class for a more detailed description of the implementation.

    '''

    asic_version = 'lightpix-v1.0'
    default_configuration_file = 'chip/default_lightpix_v1.json'
    num_registers = 239
    num_bits = 1912
    # Additional class properties regarding configuration registers are set at the end of the file.

    def __init__(self):
        # Note: properties, getters and setters are constructed after this class definition at the bottom of the file.
        super(Configuration_Lightpix_v1, self).__init__()
        return

## Set up property info
#
_property_configuration = OrderedDict([
        ('lightpix_mode',
            (v2_base._compound_property, (['lightpix_mode','hit_threshold'], (int,bool), 0, 1), (1896, 1897))),
        ('hit_threshold',
            (v2_base._compound_property, (['lightpix_mode','hit_threshold'], int, 0, 127), (1897, 1904))),
        ('timeout',
            (v2_base._basic_property, (int, 0, 255), (1904, 1912))),
    ])
_property_configuration.update(conf_v2._property_configuration)

# GENERATE THE PROPERTIES!
Configuration_Lightpix_v1.bit_map = OrderedDict()
Configuration_Lightpix_v1.register_map = OrderedDict()
Configuration_Lightpix_v1.register_names = []
v2_base._generate_properties(Configuration_Lightpix_v1, _property_configuration, verbose=False)



'''
The serial port IO interface.

'''
from __future__ import absolute_import
import time
import os
import platform
import warnings

from larpix.io import IO
from larpix import Packet
import larpix.bitarrayhelper as bah

class SerialPort(IO):
    '''Wrapper for various serial port interfaces across platforms.

       Automatically loads correct driver based on the supplied port
       name:

           - ``'/dev/anything'`` ==> Linux ==> pySerial
           - ``'scan-ftdi'`` ==> MacOS ==> libFTDI

    '''
    # Guesses for default port name by platform
    _default_port_map = {
        'Default':['/dev/ttyUSB2','/dev/ttyUSB1'], # Same as Linux
        'Linux':['/dev/serial0', '/dev/ttyAMA0', '/dev/ttyUSB2','/dev/ttyUSB1',
        '/dev/ttyUSB0'],   # Linux
        'Darwin':['scan-ftdi',],     # OS X
    }
    _logger = None
    start_byte = b'\x73'
    stop_byte = b'\x71'
    max_write = 250
    fpga_packet_size = 10
    hwm = 10000 # max bytes to read from an empty queue

    def __init__(self, port=None, baudrate=1000000, timeout=0):
        super(SerialPort, self).__init__()
        if port is None:
            port = self._guess_port()
        self.port = port
        self.resolved_port = ''
        self.port_type = ''
        self.baudrate = baudrate
        self.timeout = timeout
        self.max_write = 250
        self.serial_com = None
        self._initialize_serial_com()
        self.logger = None
        self.leftover_bytes = b''
        if not (self._logger is None):
            self.logger = self._logger
        return

    @staticmethod
    def _format_UART(packet):
        packet_bytes = packet.bytes()
        formatted_packet = (SerialPort.start_byte + packet_bytes +
                SerialPort.stop_byte)
        return formatted_packet

    @staticmethod
    def _parse_input(bytestream, leftover_bytes=b''):
        packet_size = SerialPort.fpga_packet_size
        start_byte = SerialPort.start_byte[0]
        stop_byte = SerialPort.stop_byte[0]
        data_bytes = slice(1,9)
        # parse the bytestream into Packets
        byte_packets = []
        skip_slices = []
        bytestream = leftover_bytes + bytestream
        bytestream_len = len(bytestream)
        last_possible_start = bytestream_len - packet_size
        index = 0
        while index <= last_possible_start:
            if (bytestream[index] == start_byte and
                    bytestream[index+packet_size-1] == stop_byte):
                byte_packets.append(Packet(bytestream[index+1:index+9]))
                index += packet_size
            else:
                # Throw out everything between here and the next start byte.
                # Note: start searching after byte 0 in case it's
                # already a start byte
                index = bytestream.find(start_byte, index+1)
                if index == -1:
                    index = bytestream_len
        if index <= bytestream_len:
            leftover_bytes = bytestream[index:]
        return byte_packets, leftover_bytes

    @staticmethod
    def format_bytestream(formatted_packets):
        bytestreams = []
        current_bytestream = bytes()
        for packet in formatted_packets:
            if len(current_bytestream) + len(packet) <= SerialPort.max_write:
                current_bytestream += packet
            else:
                bytestreams.append(current_bytestream)
                current_bytestream = bytes()
                current_bytestream += packet
        bytestreams.append(current_bytestream)
        return bytestreams

    @classmethod
    def encode(cls, packets):
        '''
        Encodes a list of packets into a list of bytestream messages
        '''
        return [SerialPort._format_UART(packet) for packet in packets]

    @classmethod
    def decode(cls, msgs, leftover_bytes=b''):
        '''
        Decodes a list of serial port bytestreams to packets
        '''
        packets = []
        byte_packet_list = [None] * len(msgs)
        for i in range(len(msgs)):
            byte_packet_list[i], leftover_bytes = SerialPort._parse_input(msgs[i], leftover_bytes)
        for packet_list in byte_packet_list:
            packets += packet_list
        for packet in packets:
            packet.io_channel = 1
            packet.io_group = 1
        return packets, leftover_bytes

    @classmethod
    def is_valid_chip_key(cls, key):
        '''
        Valid chip keys must be strings formatted as:
        ``'<io_chain>-<chip_id>'``

        '''
        if not super(cls, cls).is_valid_chip_key(key):
            return False
        if not isinstance(key, str):
            return False
        parsed_key = key.split('-')
        if not len(parsed_key) == 2:
            return False
        try:
            _ = int(parsed_key[0])
            _ = int(parsed_key[1])
        except ValueError:
            return False
        return True

    def send(self, packets):
        '''
        Format the packets as a bytestream and send it to the FPGA and on
        to the LArPix ASICs.

        '''
        packet_bytes = self.encode(packets)
        bytestreams = self.format_bytestream(packet_bytes)
        for bytestream in bytestreams:
            self._write(bytestream)

    def start_listening(self):
        '''
        Start listening for incoming LArPix data by opening the serial
        port.

        '''
        super(SerialPort, self).start_listening()
        self._open()

    def stop_listening(self):
        '''
        Stop listening for LArPix data by closing the serial port.

        '''
        super(SerialPort, self).stop_listening()
        self._close()

    def empty_queue(self):
        '''
        Empty the incoming data buffer and return ``(packets,
        bytestream)``.

        '''
        data_in = b''
        keep_reading = True
        count = 0
        while keep_reading:
            new_data = self._read(self.max_write)
            data_in += new_data
            count = len(data_in)
            keep_reading = (len(new_data) == self.max_write and count < self.hwm)
        packets, self.leftover_bytes = self.decode([data_in], leftover_bytes=self.leftover_bytes)
        return (packets, data_in)

    def set_larpix_uart_clk_ratio(self, value):
        '''
        Sends a special command to modify the larpix uart clk ratio (how many
        clock cycles correspond to one bit). A value of 2 means 1 uart bit == 2
        clk cycles

        '''
        data_out = (
            b'c' # start byte
            + b'\x00' # address
            + bah.fromuint(value, 8, endian='big').tobytes()
            + b'\x00'*6 # unused
            + b'q' # stop byte
            )
        self._write(data_out)

    def set_larpix_reset_cnt(self, value):
        '''
        Sends a special command to modify the length of the reset signal
        sent to the larpix chips. The reset will be held low for value + 1
        larpix clk rising edges

        '''
        data_out = (
            b'c' # start byte
            + b'\x01' # address
            + bah.fromuint(value, 8, endian='big').tobytes()
            + b'\x00'*6 # unused
            + b'q' # stop byte
            )
        self._write(data_out)

    def larpix_reset(self):
        '''
        Sends a special command to issue a larpix reset pulse. Pulse length
        is set by set_larpix_reset_cnt().

        '''
        data_out = (
            b'c' # start byte
            + b'\x02' # address
            + b'\x00'*7 # unused
            + b'q' # stop byte
            )
        self._write(data_out)

    def set_utility_pulse(self, pulse_len=None, pulse_rep=None):
        '''
        Sends a special command to issue set up utility pulser. Pulse length
        is the number of larpix clk cyles pulse is high, and pulse rep is the
        number of clk cycles until the next pulse.

        '''
        data_out = b''
        if not pulse_len is None:
            data_out += (
                b'c' # start byte
                + b'\x03' # address
                + bah.fromuint(max(pulse_len-2,0), 32, endian='big').tobytes()[::-1] # -2 for proper register value -> clk cycles conv.
                + b'\x00'*3 # unused
                + b'q' # stop byte
                )
        if not pulse_rep is None:
            data_out += (
                b'c' # start byte
                + b'\x04' # address
                + bah.fromuint(max(pulse_rep-1,0), 32, endian='big').tobytes()[::-1] # -1 for proper register value -> clk cycles conv.
                + b'\x00'*3 # unused
                + b'q' # stop byte
                )
        if data_out:
            self._write(data_out)
        else:
            raise RuntimeError('set either or both of pulse_len and pulse_rep')

    def enable_utility_pulse(self):
        '''
        Sends a special command to enable the utility pulser. Pulse
        characteristics can be set by set_utility_pulse().

        '''
        data_out = (
            b'c' # start byte
            + b'\x05' # address
            + b'\x01' # enable
            + b'\x00'*6 # unused
            + b'q' # stop byte
            )
        self._write(data_out)

    def disable_utility_pulse(self):
        '''
        Sends a special command to disable the utility pulser. Pulse
        characteristics can be set by set_utility_pulse().

        '''
        data_out = (
            b'c' # start byte
            + b'\x05' # address
            + b'\x00' # disable
            + b'\x00'*6 # unused
            + b'q' # stop byte
            )
        self._write(data_out)

    def reset(self):
        '''
        Sends a special command to reset FPGA and larpix.

        '''
        data_out = (
            b'c' # start byte
            + b'\x06' # address
            + b'\x00'*7 # unused
            + b'q' # stop byte
            )
        self._write(data_out)

    @classmethod
    def _guess_port(cls):
        '''Guess at correct port name based on platform'''
        platform_default = 'Default'
        platform_name = platform.system()
        if platform_name not in cls._default_port_map:
            platform_name = platform_default
        default_devs = cls._default_port_map[platform_name]
        osx_cmd = 'system_profiler SPUSBDataType | grep -C 7 FTDI | grep Serial'
        for default_dev in default_devs:
            if default_dev.startswith('/dev'): # pyserial
                try:
                    if os.stat(default_dev):
                        return default_dev
                except OSError:
                    continue
            elif default_dev == 'scan-ftdi':
                if platform_name == 'Darwin':  # scan for pylibftdi on OS X
                    # Scan for FTDI devices
                    result = os.popen(osx_cmd).read()
                    if len(result) > 0:
                        idx = result.find('Serial Number:')
                        dev_name = result[idx+14:idx+24].strip()
                        print('Autoscan found FTDI device: "%s"' % dev_name)
                        return dev_name
            elif not default_dev.startswith('/dev'):  # assume pylibftdi
                return default_dev
        raise OSError('Cannot find serial device for platform: %s' %
                      platform_name)

    def _ready_port(self):
        '''Function handle.  Will be reset to appropriate method'''
        raise NotImplementedError('Serial port type has not been defined.')

    def _ready_port_pyserial(self):
        '''Ready a pyserial port'''
        if not self.serial_com.is_open:
            # Open, if necessary
            self.serial_com.open()
        return

    def _ready_port_pylibftdi(self):
        '''Ready a pylibftdi port'''
        if self.serial_com.closed:
            # Open port
            self.serial_com.open()
        # Confirm baudrate (Required for OS X)
        self._confirm_baudrate()
        return

    def _ready_port_test(self):
        return True

    def _confirm_baudrate(self):
        '''Check and set the baud rate'''
        if self.serial_com.baudrate != self.baudrate:
            # Reset baudrate
            self.serial_com.baudrate = self.baudrate
        return

    def _initialize_serial_com(self):
        '''Initialize the low-level serial com connection'''
        self.resolved_port = self._resolve_port_name()
        self.port_type = self._resolve_port_type()
        if self.port_type == 'pyserial':
            self._ready_port = self._ready_port_pyserial
            import serial
            self.serial_com = serial.Serial(self.resolved_port,
                                            baudrate=self.baudrate,
                                            timeout=self.timeout)
        elif self.port_type == 'pylibftdi':
            self._ready_port = self._ready_port_pylibftdi
            import pylibftdi
            self.serial_com = pylibftdi.Device(self.resolved_port)
        elif self.port_type == 'test':
            self._ready_port = self._ready_port_test
            import test.test_larpix as test_lib
            self.serial_com = test_lib.FakeSerialPort()
        else:
            raise ValueError('Port type must be either pyserial, pylibftdi, or test')
        return

    def _resolve_port_name(self):
        '''Resolve the serial port name, based on user request'''
        if self.port is None:
            # Must set port
            raise ValueError('You must choose a serial port for operation')
        if self.port == 'auto':
            # Try to guess the correct port
            return self._guess_port()
        return self.port

    def _resolve_port_type(self):
        '''Resolve the type of serial port, based on the name'''
        if isinstance(self.resolved_port, str):
            if self.resolved_port.startswith('/dev'):
                # Looks like a tty device.  Use pyserial.
                return 'pyserial'
            elif self.resolved_port == 'test':
                # Testing port. Don't use an external library
                return 'test'
            elif not self.resolved_port.startswith('/dev'):
                # Looks like a libftdi raw device.  Use pylibftdi.
                return 'pylibftdi'
        raise ValueError('Unknown port: %s' % self.port)

    def _open(self):
        '''Open the port'''
        self._ready_port()
        return

    def _close(self):
        '''Close the port'''
        if self.serial_com is None: return
        self.serial_com.close()

    def _write(self, data):
        '''Write data to serial port'''
        self._ready_port()
        write_time = time.time()
        self.serial_com.write(data)
        if not self.is_listening:
            self._close()
        if self.logger:
            self.logger.record({'data_type':'write','data':data,'time':write_time})
        return

    def _read(self, nbytes):
        '''Read data from serial port'''
        self._ready_port()
        read_time = time.time()
        data = self.serial_com.read(nbytes)
        if self.logger:
            self.logger.record({'data_type':'read','data':data,'time':read_time})
        return data

def enable_logger(filename=None):
    '''Enable serial data logger'''
    if SerialPort._logger is None:
        from larpix.serial_helpers.datalogger import DataLogger
        SerialPort._logger = DataLogger(filename)
    if not SerialPort._logger.is_enabled():
        SerialPort._logger.enable()
    return

def disable_logger():
    '''Disable serial data logger'''
    if SerialPort._logger is not None:
        SerialPort._logger.disable()
    return

def flush_logger():
    '''Flush serial data logger data to output file'''
    if SerialPort._logger is not None:
        SerialPort._logger.flush()
    return

def _test_serial_loopback(port_name='auto', enable_logging=False):
    '''Write stream of integers to serial port.  Read back and see if
    loopback data is correct.'''
    baudrate = 1000000
    timeout=0.1
    if enable_logging:
        enable_logger()
    serial_port = SerialPort(port_name)
    serial_port.baudrate = baudrate
    print(' serial baudrate:',serial_port.baudrate)
    serial_port._open()
    test_length = 256
    n_errors = 0
    max_read_length = 8192
    for iter_idx in range(10):
        write_data = range(iter_idx*10,iter_idx*10+test_length)
        write_data = [elem % 256 for elem in write_data]
        write_bits = bytearray(write_data)
        serial_port._write(write_bits)
        read_bits = b''
        read_bits += serial_port._read(max_read_length)
        print("Testing:" + str([write_bits,]))
        if str(write_bits) != str(read_bits):
            print(" Error:")
            print("  wrote: ", str(write_bits))
            print("   read: ", str(read_bits))
            print("   read_bytes / wrote_bytes: %d / %d " % (len(read_bits),
                                                             test_length))
            n_errors += 1
        else:
            print(' OK')
    serial_port._close()
    return n_errors

if '__main__' == __name__:
    _test_serial_loopback()




from larpix import Configuration_Lightpix_v1
from bitarray import bitarray
import larpix.bitarrayhelper as bah
from larpix import configs
import json

def test_conf_all_registers():
    default_filename = 'chip/default_lightpix_v1.json'
    data = configs.load(default_filename)

    c = Configuration_Lightpix_v1()
    print(data, len(data))
    print(c.register_names,len(c.register_names))
    # check that all registers are in the configuration file
    assert set(data['register_values'].keys()) == set(c.register_names)
    for register in c.register_names:
        print('testing {}'.format(register))
        config_value = data['register_values'][register]
        stored_value = getattr(c,register)

        # test getters
        assert stored_value == config_value

        # test setters
        if register == 'chip_id':
            new_value = 12
        elif isinstance(stored_value, list):
            new_value = []
            for value in stored_value:
                if value == 1:
                    new_value += [0]
                elif value == 0:
                    new_value += [1]
                else:
                    new_value += [value-1]
        elif isinstance(stored_value, int):
            if stored_value == 1:
                new_value = 0
            elif stored_value == 0:
                new_value = 1
            else:
                new_value = stored_value-1
        setattr(c,register,new_value)
        assert getattr(c,register) == new_value

        # test data getters
        assert list(c.register_map[register]) == [reg for reg,bits in getattr(c,register+'_data')]

def test_conf():
    c = Configuration_Lightpix_v1()
    endian = 'little'

    # test simple register
    c.timeout = 12
    assert c.timeout == 12
    assert c.timeout_data == [(238, bah.fromuint(12,8, endian=endian))]

    c.timeout_data = (238, bah.fromuint(90,8, endian=endian))
    assert c.timeout == 90
    assert c.timeout_data == [(238, bah.fromuint(90,8, endian=endian))]

    c.timeout_data = bah.fromuint(34,8, endian=endian)
    assert c.timeout == 34
    assert c.timeout_data == [(238, bah.fromuint(34,8, endian=endian))]

    # test compound register
    c.lightpix_mode = 1
    c.hit_threshold = 3
    reg_data = [(237, bitarray('11100000'))]
    assert c.lightpix_mode == 1
    assert c.lightpix_mode_data == reg_data
    assert c.hit_threshold == 3
    assert c.hit_threshold_data == reg_data

    c.hit_threshold = 0
    reg_data = [(237, bitarray('10000000'))]
    assert c.lightpix_mode == 1
    assert c.lightpix_mode_data == reg_data
    assert c.hit_threshold == 0
    assert c.hit_threshold_data == reg_data

def test_compare():
    c = Configuration_Lightpix_v1()
    other = Configuration_Lightpix_v1()
    assert c.compare(other) == {}

    c.lightpix_mode = 1
    other.lightpix_mode = 0
    assert c.compare(other) == {'lightpix_mode': (1,0)}

def test_get_nondefault_registers():
    c = Configuration_Lightpix_v1()
    c.lightpix_mode = 1
    c.hit_threshold = 8
    c.timeout = 34
    assert c.get_nondefault_registers() == {
            'lightpix_mode': (1,0),
            'hit_threshold': (8,16),
            'timeout': (34, 30)
            }



'''
This is an alternative to the ``larpix.format.hdf5format``format that allows for
much faster conversion to file at the expense of human readability.

To use, pass a list of bytestring messages into the ``to_rawfile()`` method::

    msgs = [b'this is a test message', b'this is a different message']
    to_rawfile('raw.h5', msgs)

To access the data in the file, the inverse method ``from_rawfile()`` is used::

    rd = from_rawfile('raw.h5')
    rd['msgs'] # [b'this is a test message', b'this is a different message']

Messages may be recieved from multiple ``io_group`` sources, in this case, a
per-message header with ``io_group`` can be specified as a list of integers of
the same length as the ``msgs`` list and passed into the file at the same time::

    msgs = [b'message from 1', b'message from 2']
    io_groups = [1, 2]
    to_rawfile('raw.h5', msgs=msgs, msg_headers={'io_groups': io_groups})

    rd = from_rawfile('raw.h5')
    rd['msgs'] # [b'message from 1', b'message from 2']
    rd['msg_headers']['io_groups'] # [1, 2]

File versioning
---------------

Some version validation is included with the file format through
the ``version`` and ``io_version`` file metadata. When creating a new file, a
file format version can be provided with the ``version`` keyword argument as
a string formatted ``'major.minor'``::

    to_rawfile('raw_v0_0.h5', version='0.0')

Subsequent writes to the file will only occur if the requested file version and
the existing file versions are compatible. Incompatiblity occurs if there is
a difference in the major version number or the minor version number is less
than the requested file version::

    to_rawfile('raw_v0_0.h5', version='0.1') # fails due to minor version incompatibility
    to_rawfile('raw_v0_0.h5', version='1.0') # fails due to major version incompatibility

By default, the most recent file version is used.

On the file read side, a version number can be requested and the file will be
parsed assuming a specific version::

    from_rawfile('raw_v0_0.h5', version='0.0')
    from_rawfile('raw_v0_0.h5', version='0.1') # fails due to minor version incompatiblity
    from_rawfile('raw_v0_0.h5', version='1.0') # fails due to major version compatibility

The ``io_version`` optional metadata marks the version of the io message format
that was used to encode the message bytestrings. If present as a keyword
argument when writing to the file, an ``AssertionError`` will be raised if
the io version is incompatible with the existing one stored in metadata::

    to_rawfile('raw_io_v0_0.h5', io_version='0.0')
    to_rawfile('raw_io_v0_0.h5', io_version='0.1') # fails due to minor version incompatibility
    to_rawfile('raw_io_v0_0.h5', io_version='1.0') # fails due to major version incompatibility

A similar mechanism occurs when requesting an io version when reading from the
file::

    from_rawfile('raw_io_v0_0.h5', io_version='0.1') # fails due to minor version incompatibility
    from_rawfile('raw_io_v0_0.h5', io_version='1.0') # fails due to major version
    from_rawfile('raw_io_v0_0.h5', io_version='0.0')

I think it is worthwhile to further clarify the ``io_version`` and the file
``version``, as this might be confusing. In particular, you might be asking,
"What io versions are compatible with what file versions?" The ``rawhdf5format``
is a way of wrapping raw binary data into a format that only requires HDF5 to
parse. The file version represents this HDF5 structuring (the hdf5 dataset
formats, file metadata, what message header data is available). Whereas the
``io_version`` represents the formatting of the binary data that the file
contains. So the answer to that question is: *all* file versions are compatible
with *all* io versions.

Converting to other file types
------------------------------

This format was created with a specific application in mind - provide a
temporary but fast file format for PACMAN messages. When used in this
case, to convert to the standard ``larpix.format.hdf5format``::

    from larpix.format.pacman_msg_format import parse
    from larpix.format.hdf5format import to_file

    rd = from_rawfile('raw.h5')
    pkts = list()
    for io_group,msg in zip(rd['msg_headers']['io_groups'], rd['msgs']):
        pkts.extend(parse(msg, io_group=io_group))
    to_file('new_filename.h5', packet_list=pkts)

but as always, the most efficient means of accessing the data is to operate on
the data itself, rather than converting between types.

Metadata (v0.0)
---------------
The group ``meta`` contains file metadata stored as attributes:

    - ``created``: ``float``, unix timestamp since the 1970 epoch in seconds indicating when file was first created

    - ``modified``: ``float``, unix timestamp since the 1970 epoch in seconds indicating when the file was last written to

    - ``version``: ``str``, file version, formatted as ``'major.minor'``

    - ``io_version``: ``str``, optional version for message bytestring encoding, formatted as ``'major.minor'``

Datasets (v0.0)
---------------
The hdf5 format contains two datasets ``msgs`` and ``msg_headers``:

    - ``msgs``: shape ``(N,)``; variable-length ``uint1`` arrays encoding each message bytestring

    - ``msg_headers``: shape ``(N,)``; numpy structured array with fields:

        - ``'io_group'``: ``uint1`` representing the ``io_group`` associated with each message

'''
import time
import warnings
import os
os.environ['HDF5_USE_FILE_LOCKING'] = 'FALSE' # needed for error-free SWMR writer access
_file_read_reattempts = 100 # also needed to ignore inevitable reader errors for SWMR mode

import h5py
import numpy as np

#: Most up-to-date raw larpix hdf5 format version.
latest_version = '0.0'

#: Description of the datasets and their dtypes used in each version of the raw larpix hdf5 format.
#:
#: Structured as ``dataset_dtypes['<version>']['<dataset>'] = <dtype>``.
dataset_dtypes = {
    '0.0': {
        'msgs': h5py.vlen_dtype(np.dtype('u1')),
        'msg_headers': np.dtype([
            ('io_groups','u1')
            ])
    }
}
def _store_msgs_v0_0(msgs, version):
    msg_dtype = np.dtype('u1')
    arr_dtype = dataset_dtypes[version]['msgs']
    return np.array([np.frombuffer(msg, dtype=msg_dtype) for msg in msgs], dtype=arr_dtype)

def _store_msg_headers_v0_0(msg_headers, version):
    length = len(msg_headers['io_groups'])
    arr = np.zeros((length,), dtype=dataset_dtypes[version]['msg_headers'])
    for key in msg_headers:
        arr[key] = msg_headers[key]
    return arr

def _parse_msgs_v0_0(msgs, version):
    return [msg.tobytes() for msg in msgs]

def _parse_msg_headers_v0_0(msg_headers, version):
    rd = dict()
    for key in msg_headers.dtype.names:
        rd[key] = list(msg_headers[key].astype(int))
    return rd

def _store_msgs(msgs, version):
    '''
    A version-safe way to put messages into the dataset

    :param msgs: an iterable of PACMAN messages to convert

    :param version: version string

    :returns: a numpy array, 1 row for each msg

    '''
    return _store_msgs_v0_0(msgs, version)

def _store_msg_headers(msg_headers, version):
    '''
    A version-safe way to put message headers into the dataset

    :param msg_headers: a dict of iterable values for each field in ``dataset_dtypes[version]['msg_headers'].names`` to convert to a numpy array

    :param version: version string

    :returns: a numpy array, 1 row for each io_group

    '''
    return _store_msg_headers_v0_0(msg_headers, version)

def _parse_msgs(msgs, version):
    '''
    A version-safe conversion of numpy array void objects into PACMAN message byte strings

    :param msgs: a list of void-type numpy arrays

    :param version: version string

    :returns: list of PACMAN message byte strings, 1 for each row in data
    '''
    return _parse_msgs_v0_0(msgs, version)

def _parse_msg_headers(msg_headers, version):
    '''
    A version-safe conversion of message header numpy arrays into lists

    :param msg_headers: a structure array of the msg_header dataset to convert to a list of values

    :param version: version string

    :returns: dict of lists, keyed by dtype fields, 1 entry in each list for each row in data
    '''
    return _parse_msg_headers_v0_0(msg_headers, version)

def to_rawfile(filename, msgs=None, version=None, msg_headers=None, io_version=None):
    '''
    Write a list of bytestring messages to an hdf5 file. If the file exists,
    the messages will appended to the end of the dataset.

    :param filename: desired filename for the file to write or update

    :param msgs: iterable of variable-length bytestrings to write to the file. If ``None`` specified, will only create file and update metadata.

    :param version: a string of major.minor version desired. If ``None`` specified, will use the latest file format version (if new file) or version in file (if updating an existing file).

    :param msg_headers: a dict of iterables to associate with each message header. Iterables must be same length as ``msgs``. If ``None`` specified, will use a default value of ``0`` for each message. Keys are dtype field names specified in ``dataset_dtypes[version]['msg_headers'].names``

    :param io_version: optional metadata to associate with file corresponding to the io format version of the bytestring messages. Throws ``RuntimeError`` if version incompatibility encountered in an existing file.

    '''
    now = time.time()
    with h5py.File(filename, 'a', libver='latest') as f:
        if 'meta' not in f.keys():
            # new file
            version = latest_version if version is None else version
            f.create_group('meta')
            f['meta'].attrs['version'] = version
            f['meta'].attrs['created'] = now
            f['meta'].attrs['modified'] = now
            if io_version is not None:
                f['meta'].attrs['io_version'] = io_version

            # get current position in file
            curr_idx = 0

            # create datasets
            f.create_dataset('msgs', shape=(0,), maxshape=(None,), compression='gzip', dtype=dataset_dtypes[version]['msgs'])
            f.create_dataset('msg_headers', shape=(0,), maxshape=(None,), compression='gzip', dtype=dataset_dtypes[version]['msg_headers'])

            f.swmr_mode = True
        else:
            # existing file
            f.swmr_mode = True

            file_version = f['meta'].attrs['version']
            assert (file_version == version) or (version is None), \
                'Version mismatch! file: {}, requested: {}'.format(file_version, version)
            version = file_version
            assert (io_version is None) or ('io_version' in f['meta'].attrs.keys() and f['meta'].attrs['io_version'].split('.')[0] == io_version.split('.')[0] and f['meta'].attrs['io_version'].split('.')[-1] >= io_version.split('.')[-1]), 'IO version mismatch! file: {}, requested {}'.format(f['meta'].attrs['io_version'], io_version)

            f['meta'].attrs['modified'] = now

        # update data
        if msgs is not None:
            headers = dict()
            for key in msg_headers:
                headers[key] = msg_headers[key]
                if key not in dataset_dtypes[version]['msg_headers'].names:
                    raise RuntimeError('Encountered unknown message header key {}'.format(key))
            for key in dataset_dtypes[version]['msg_headers'].names:
                if key not in headers:
                    headers[key] = np.zeros(len(msgs))
                assert len(headers[key]) == len(msgs), 'Data length mismatch! msgs is length {}, but msg_headers field {} is length {}'.format(len(msgs),key,len(headers[key]))

            # resize datasets
            curr_idx = len(f['msgs'])
            f['msgs'].resize((curr_idx+len(msgs),))
            f['msg_headers'].resize((curr_idx+len(msgs),))

            # store in file
            msgs_array = _store_msgs(
                msgs,
                version=version
                )
            msg_headers_array = _store_msg_headers(
                msg_headers,
                version=version
                )

            f['msgs'][curr_idx:curr_idx + len(msgs_array)] = msgs_array
            f['msg_headers'][curr_idx:curr_idx + len(msg_headers_array)] = msg_headers_array

        # flush
        f['msgs'].flush()
        f['msg_headers'].flush()

def _synchronize(attempts, *dsets):
    if len(dsets) <= 1:
        return
    success = attempts == 0
    attempt = 1
    lengths = [0]*len(dsets)
    while (attempt <= attempts or attempts < 0) and not success:
        attempt += 1
        for i,dset in enumerate(dsets):
            dset.id.refresh()
            lengths[i] = len(dset)
        if all([lengths[0] == length for length in lengths[1:]]):
            success = True
    if not success:
        raise RuntimeError('Could not achieve a stable file state after {} attempts!'.format(attempts))

def len_rawfile(filename, attempts=1):
    '''
    Check the total number of messages in a file

    :param filename: filename to check

    :param attempts: a parameter only relevant if file is being actively written to by another process, specifies number of refreshes to try if a synchronized state between the datasets is not achieved. A value less than ``0`` busy blocks until a synchronized state is achieved. A value greater than ``0`` tries to achieve synchronization a max of ``attempts`` before throwing a ``RuntimeError``. And a value of ``0`` does not attempt to synchronize (not recommended).

    :returns: ``int`` number of messages in file

    '''
    err = None
    for _ in range(_file_read_reattempts):
        try:
            with h5py.File(filename, 'r', swmr=True, libver='latest') as f:
                _synchronize(attempts, f['msgs'], f['msg_headers'])
                return len(f['msgs'])
        except OSError as e:
            if e.errno is None:
                warnings.warn(str(e) + '\ntrying again...', RuntimeWarning)
                err = e
            else:
                raise e
    raise err

def from_rawfile(filename, start=None, end=None, version=None, io_version=None, msg_headers_only=False, mask=None, attempts=1):
    '''
    Read a chunk of bytestring messages from an existing file

    :param filename: filename to read bytestrings from

    :param start: index for the start position when reading from the file (default = ``None``). If a value less than 0 is specified, index is relative to the end of the file. If ``None`` is specified, data is read from the start of the file. If a ``mask`` is specified, does nothing.

    :param end: index for the end position when reading from the file (default = ``None``). If a value less than 0 is specified, index is relative to the end of the file. If ``None`` is specified, data is read until the end of the file. If a ``mask`` is specified, does nothing.

    :param version: required version compatibility. If ``None`` specified, uses the version stored in the file metadata

    :param io_version: required io version compatibility. If ``None`` specified, does not check the ``io_version`` file metadata

    :param msg_headers_only: optional flag to only load header information and not message bytestrings (``'msgs'`` value in return dict will be ``None`` if ``msg_headers_only=True``)

    :param mask: boolean mask alternative to ``start`` and ``end`` chunk specification to indicate specific file rows to load. Boolean 1D array with length equal to ``len_rawfile(filename)``

    :param attempts: a parameter only relevant if file is being actively written to by another process, specifies number of refreshes to try if a synchronized state between the datasets is not achieved. A value less than ``0`` busy blocks until a synchronized state is achieved. A value greater than ``0`` tries to achieve synchronization a max of ``attempts`` before throwing a ``RuntimeError``. And a value of ``0`` does not attempt to synchronize (not recommended).

    :returns: ``dict`` with keys for ``'created'``, ``'modified'``, ``'version'``, and ``'io_version'`` metadata, along with ``'msgs'`` (a ``list`` of bytestring messages) and ``'msg_headers'`` (a dict with message header field name: ``list`` of message header field data, 1 per message)

    '''
    err = None
    for _ in range(_file_read_reattempts):
        try:
            with h5py.File(filename, 'r', swmr=True, libver='latest') as f:
                # fetch metadata
                created = f['meta'].attrs['created']
                modified = f['meta'].attrs['modified']

                # check file format version is compatible
                file_version = f['meta'].attrs['version']
                version_major = file_version.split('.')[0]
                assert (version is None) or (file_version >= version and version_major == version.split('.')[0]), 'Incompatible version mismatch! file: {}, requested: {}'.format(file_version, version)
                version_minor = min(file_version.split('.')[-1], version.split('.')[-1]) if version is not None else file_version.split('.')[-1]
                version = '{}.{}'.format(version_major,version_minor)

                # check io format version is compatible
                file_io_version = f['meta'].attrs['io_version'] if 'io_version' in f['meta'].attrs.keys() else None
                io_version_major, io_version_minor = file_io_version.split('.') if file_io_version is not None else (None,None)
                assert (io_version is None) or (file_io_version is None) or (io_version_major == io_version.split('.')[0] and io_version_minor >= io_version.split('.')[-1]), 'IO version mismatch! file: {}, requested {}'.format(file_io_version,io_version)
                io_version = file_io_version

                # check to make sure that the msgs and headers dsets are synchronized
                _synchronize(attempts, f['msgs'], f['msg_headers'])

                # define chunk of data to load
                start = int(start) if start is not None else 0
                end = int(end) if end is not None else len(f['msgs'])
                mask = mask if mask is not None else slice(start,end)

                # get data from file
                msg_headers = _parse_msg_headers(f['msg_headers'][mask], version)
                msgs = _parse_msgs(f['msgs'][mask], version) if not msg_headers_only else None

                return dict(
                    created=created,
                    modified=modified,
                    version=version,
                    io_version=io_version,
                    msgs=msgs,
                    msg_headers=msg_headers
                    )
        except OSError as e:
            if e.errno is None:
                warnings.warn(str(e) + '\ntrying again...', RuntimeWarning)
                err = e
            else:
                raise e
        raise err




import itertools
import zmq
import bidict
import time
from collections import defaultdict
import multiprocessing
import sys
if sys.version_info[0] >= 3:
    from queue import Empty
else:
    from Queue import Empty
import os

from larpix.io import IO
from larpix.configs import load
import larpix.format.pacman_msg_format as pacman_msg_format
import larpix.format.rawhdf5format as rawhdf5format
from larpix import Packet_v2

class PACMAN_IO(IO):
    '''
    The PACMAN_IO object interfaces with a network of PACMAN
    boards each running a pacman-cmdserver and pacman-dataserver.

    This object handles the ZMQ messaging protocol to send and receive
    formatted messages to/from the PACMAN boards. If you want more
    info on how messages are formatted, see ``larpix.format.pacman_msg_format``.

    The PACMAN_IO object has five flags for optimizing communications
    which you may or may not want to enable:

        - ``group_packets_by_io_group``
        - ``interleave_packets_by_io_channel``
        - ``double_send_packets``
        - ``enable_raw_file_writing``
        - ``disable_packet_parsing``

    To enable each option set the flag to ``True``; to disable, set to
    ``False``.

        - The ``group_packets_by_io_group`` option is enabled by default and assembles the packets sent in each call to ``send`` into as few messages as possible destined for a single io group. E.g. sending three packets (2 for ``io_group=1``, 1 for ``io_group=2``) will combine the packets for ``io_group=1`` into a single message to transfer over the network. This reduces overhead associated with network latency and allows large data transfers to happen much faster.

        - The ``interleave_packets_by_io_channel`` option is enabled by default and interleaves packets within a given message to each io_channel on a given io_group. E.g. 3 packets destined for ``io_channel=1``, ``io_channel=1``, and ``io_channel=2`` will be reordered to ``io_channel=1``, ``io_channel=2``, and ``io_channel=1``. The order of the packets is preserved for each io_channel. This increases the data throughput by about a factor of N, where N is the number of io channels in the message.

        - The ``double_send_packets`` option is disabled by default and duplicates each packet sent to the PACMAN by a call to ``send()``. This is potentially useful for working around the 512 bug when you need to insure that a packet reaches a chip, but you don't care about introducing extra packets into the system (i.e. when configuring chips).

        - The ``enable_raw_file_writing`` option will directly dump data to a larpix raw hdf5 formatted file. This is used as a more performant means of logging data (see ``larpix.format.rawhdf5format``). The data file name can be accessed or changed via the ``raw_filename`` attribute, or can be set when creating the ``PACMAN_IO`` object with the ``raw_directory`` and ``raw_filename`` keyword args.

        - The ``disable_packet_parsing`` option will skip converting PACMAN messages into ``larpix.packet`` types. Thus if ``disable_packet_parsing=True``, every call to ``empty_queue`` will return ``[], b''``. Typically used in conjunction with ``enable_raw_file_writing``, this allows the PACMAN_IO class to read data much faster.


    '''
    default_filepath = 'io/pacman.json'
    default_raw_filename_fmt = 'raw_%Y_%m_%d_%H_%M_%S_%Z.h5'
    max_msg_length = 2**16-1
    cmdserver_port = '5555'
    dataserver_port = '5556'
    _valid_config_classes = ['PACMAN_IO']

    group_packets_by_io_group = True
    interleave_packets_by_io_channel = True
    double_send_packets = False
    enable_raw_file_writing = False
    disable_packet_parsing = False

    _base_ctrl_reg = 0x10
    _clk_ctrl_reg = 0x1010
    _sw_reset_cycles_reg = 0x1014
    _channel_offset = 0x2000
    _channel_size = 0x1000
    _uart_clock_ratio_offset = 0x10
    _vddd_dac_reg = 0x24001
    _vdda_dac_reg = 0x24011
    _vddd_adc_reg = 0x24032
    _vdda_adc_reg = 0x24042
    _iddd_adc_reg = 0x24031
    _idda_adc_reg = 0x24041
    _vplus_adc_reg = 0x24022
    _iplus_adc_reg = 0x24021
    _adc2mv = lambda _,x: ((x >> 16) >> 3) * 4
    _adc2ma = lambda _,x: ((x >> 16) - (x >> 31) * 65535) * 500 * 0.01

    def __init__(self, config_filepath=None, hwm=20000, relaxed=True, timeout=-1, raw_directory='./', raw_filename=None):
        super(PACMAN_IO, self).__init__()
        self.load(config_filepath)

        self.context = zmq.Context()
        self.senders = bidict.bidict()
        self.receivers = bidict.bidict()
        for address in self._io_group_table.inv:
            self.senders[address] = self.context.socket(zmq.REQ)
            self.receivers[address] = self.context.socket(zmq.SUB)
        self.hwm = hwm
        for receiver in self.receivers.values():
            receiver.set_hwm(self.hwm)
            receiver.setsockopt(zmq.CONNECT_TIMEOUT,max(timeout,0))
            receiver.setsockopt(zmq.LINGER,0)
            receiver.setsockopt(zmq.RCVTIMEO,timeout)
        for sender in self.senders.values():
            if relaxed:
                sender.setsockopt(zmq.REQ_RELAXED,True)
            sender.setsockopt(zmq.LINGER,0)
            sender.setsockopt(zmq.CONNECT_TIMEOUT,max(timeout,0))
            sender.setsockopt(zmq.RCVTIMEO,timeout)
            sender.setsockopt(zmq.SNDTIMEO,timeout)
        for address in self._io_group_table.inv:
            send_address = 'tcp://' + address + ':' + self.cmdserver_port
            receive_address = 'tcp://' + address + ':' + self.dataserver_port
            self.senders[address].connect(send_address)
            self.receivers[address].connect(receive_address)
        self._sender_replies = defaultdict(list)
        self.poller = zmq.Poller()
        for receiver in self.receivers.values():
            self.poller.register(receiver, zmq.POLLIN)

        self._raw_file_queue = multiprocessing.Queue()
        self.raw_filename = os.path.join(
            raw_directory,
            raw_filename if raw_filename is not None \
                else time.strftime(self.default_raw_filename_fmt)
        )
        self._launch_raw_file_worker()

    def send(self, packets):
        '''
        Sends a request message to PACMAN boards to send designated
        packets.

        '''
        msg_packets = list()
        # group packets into messages destined for a single io group (otherwise 1pkt = 1msg)
        if self.group_packets_by_io_group:
            grouped_packets = self._group_by_attr(packets, 'io_group')
            for io_group, packets in grouped_packets.items():
                msg_packets.append(packets)
        else:
            for packet in packets:
                msg_packets.append([packet])

        # interleave across io group channels
        if self.interleave_packets_by_io_channel and self.group_packets_by_io_group:
            interleaved_msg_packets = list()
            for packets in msg_packets:
                interleaved_msg_packets.append(self._interleave_by_attr(packets, 'io_channel'))
            msg_packets = interleaved_msg_packets

        # double up sent packets to help avoid 512 bug
        if self.double_send_packets:
            doubled_msg_packets = list()
            for packets in msg_packets:
                for _ in range(2):
                    doubled_msg_packets.append(packets)
            msg_packets = doubled_msg_packets

        # convert packets to messages
        resp_addresses = list()
        for packets in msg_packets:
            io_group = packets[0].io_group
            for i in range(0, len(packets), self.max_msg_length):
                #for packet in packets: print(packet)
                msg_len = min(len(packets)-i, self.max_msg_length)
                msg = pacman_msg_format.format(packets[i:i+msg_len], msg_type='REQ')
                address = self._io_group_table[io_group]
                self.senders[address].send(msg)
                self._sender_replies[address].append(self.senders[address].recv())

    def start_listening(self):
        '''
        Start keeping msgs from data server

        '''
        if self.is_listening:
            raise RuntimeError('Already listening')
        super(PACMAN_IO, self).start_listening()
        for receiver in self.receivers.values():
            receiver.setsockopt(zmq.SUBSCRIBE, b'')

    def stop_listening(self):
        '''
        Stop keeping msgs from data server

        '''
        if not self.is_listening:
            raise RuntimeError('Already not listening')
        super(PACMAN_IO, self).stop_listening()
        for receiver in self.receivers.values():
            receiver.setsockopt(zmq.UNSUBSCRIBE, b'')

    @staticmethod
    def _group_by_attr(packets, attr):
        '''
        Groups packets by the specified attribute
        returns a dict of attr_value: [<packets w/ attr=attr_value>]

        '''
        groupings = defaultdict(list)
        for packet in packets:
            groupings[getattr(packet,attr)].append(packet)
        return groupings

    @staticmethod
    def _interleave_by_attr(packets, attr):
        '''
        Interleaves packets by the specified attribute, e.g. by io channel::

            in_packets = [<ch 1>, <ch 1>, ..., <ch 2>, <ch 2>, ..]
            _interleave_by_attr(in_packets, 'io_channel')
            # returns [<ch 1>, <ch 2>, <ch 1>, <ch 2>, ... , <ch 1>, <ch 1>]

        '''

        groupings = PACMAN_IO._group_by_attr(packets, attr)
        zipped_packets = itertools.zip_longest(*groupings.values(), fillvalue=None)
        interleaved = list()
        for row in zipped_packets:
            for packet in row:
                if packet is not None:
                    interleaved.append(packet)
        return interleaved

    def empty_queue(self):
        '''
        Fetch and parse waiting packets on pacman data socket

        returns tuple of list of packets, full bytestream of all messages

        '''
        packets = []
        address_list = list()
        bytestream_list = list()
        bytestream = b''
        n_recv = 0
        while self.poller.poll(0) and n_recv < self.hwm:
            events = dict(self.poller.poll(0))
            for socket, n_events in events.items():
                for _ in range(n_events):
                    message = socket.recv()
                    n_recv += 1
                    bytestream_list += [message]
                    address_list += [self.receivers.inv[socket]]
        if not self.disable_packet_parsing:
            for message, address in zip(bytestream_list, address_list):
                packets += pacman_msg_format.parse(message, io_group=self._io_group_table.inv[address])
            bytestream = b''.join(bytestream_list)
        if self.enable_raw_file_writing:
            self._raw_file_queue.put((bytestream_list, [self._io_group_table.inv[address] for address in address_list]))
            if not self._raw_file_worker.is_alive():
                self._launch_raw_file_worker()

        return packets,bytestream

    def cleanup(self):
        '''
        Close the ZMQ objects to prevent a memory leak.

        This method is only required if you plan on instantiating a new
        ``PACMAN_IO`` object.

        '''
        for address in self.senders.keys():
            self.senders[address].close(linger=0)
            self.receivers[address].close(linger=0)
        self.context.term()

    @staticmethod
    def _to_raw_file(queue_, filename, timeout=1, max_msgs=100000):
        start_time = time.time()
        while (time.time() < start_time + timeout or not queue_.empty()):
            # wait for data
            try:
                msgs, io_groups = queue_.get(timeout=timeout)
            except Empty:
                continue
            # buffer data
            while len(msgs) < max_msgs:
                try:
                    new_msgs, new_io_groups = queue_.get(False)
                    msgs.extend(new_msgs)
                    io_groups.extend(new_io_groups)
                except Empty:
                    break
            # write to file
            if len(msgs):
                rawhdf5format.to_rawfile(filename, msgs=msgs, msg_headers={'io_groups': io_groups}, io_version=pacman_msg_format.latest_version)
                start_time = time.time()

    def _launch_raw_file_worker(self):
        self._raw_file_worker = multiprocessing.Process(target=self._to_raw_file, args=(self._raw_file_queue, self.raw_filename))
        self._raw_file_worker.start()

    def join(self):
        '''
        Wait for raw file worker to finish

        '''
        self._raw_file_worker.join()

    @property
    def raw_filename(self):
        return self._raw_filename

    @raw_filename.setter
    def raw_filename(self,value):
        if hasattr(self,'_raw_filename') \
                and value != self._raw_filename \
                and self._raw_file_worker.is_alive():
            self.join()
        self._raw_filename = value

    def set_reg(self, reg, val, io_group=None):
        '''
        Set a 32-bit register in the pacman PL

        '''
        if io_group is None:
            return dict([(io_group,self.set_reg(reg, val, io_group=io_group)) for io_group in self._io_group_table])
        msg = pacman_msg_format.format_msg('REQ',[('WRITE',reg,val)])
        addr = self._io_group_table[io_group]
        self.senders[addr].send(msg)
        self._sender_replies[addr].append(self.senders[addr].recv())

    def get_reg(self, reg, io_group=None):
        '''
        Read a 32-bit register from the pacman PL

        If no ``io_group`` is specified, returns a ``dict`` of ``io_group, reg_value``
        else returns reg_value

        '''
        if io_group is None:
            return dict([(io_group, self.get_reg(reg, io_group=io_group)) for io_group in self._io_group_table])
        msg = pacman_msg_format.format_msg('REQ',[('READ',reg,0)])
        addr = self._io_group_table[io_group]
        self.senders[addr].send(msg)
        self._sender_replies[addr].append(self.senders[addr].recv())
        msg_data = pacman_msg_format.parse_msg(self._sender_replies[addr][-1])
        if msg_data[1][0][0] == 'READ':
            return msg_data[1][0][-1]
        raise RuntimeError('Error received from server')

    def ping(self, io_group=None):
        '''
        Send a ping message

        If no ``io_group`` is specified, returns a ``dict`` of ``io_group, response``
        else returns response

        '''
        if io_group is None:
            return dict([(io_group, self.ping(io_group=io_group)) for io_group in self._io_group_table])
        msg = pacman_msg_format.format_msg('REQ',[('PING',)])
        addr = self._io_group_table[io_group]
        try:
            self.senders[addr].send(msg)
            self._sender_replies[addr].append(self.senders[addr].recv())
            msg_data = pacman_msg_format.parse_msg(self._sender_replies[addr][-1])
            if msg_data[1][0][0] == 'PONG':
                return True
        except zmq.ZMQError as e:
            print('IO error on {}: {}'.format(io_group,e))
        return False

    def get_vddd(self, io_group=None):
        '''
        Gets PACMAN VDDD voltage

        Returns VDDD and IDDD values from the built-in ADC as
        a tuple of mV and mA respectively

        '''
        if io_group is None:
            return dict([(io_group, self.get_vddd(io_group=io_group)) for io_group in self._io_group_table])
        mv = self._adc2mv(self.get_reg(self._vddd_adc_reg, io_group=io_group))
        ma = self._adc2ma(self.get_reg(self._iddd_adc_reg, io_group=io_group))
        return mv, ma

    def set_vddd(self, vddd_dac=0xD5A3, io_group=None, settling_time=0.1):
        '''
        Sets PACMAN VDDD voltage

        If no ``vddd_dac`` value is specified, sets VDDD to default of ~1.8V

        Returns the resulting VDDD and IDDD values from the built-in ADC as
        a tuple of mV and mA respectively

        '''
        if io_group is None:
            return dict([(io_group, self.set_vddd(vddd_dac, io_group=io_group)) for io_group in self._io_group_table])
        self.set_reg(self._vddd_dac_reg, vddd_dac, io_group=io_group)
        if settling_time:
            time.sleep(settling_time)
        return self.get_vddd(io_group=io_group)

    def get_vdda(self, io_group=None):
        '''
        Gets PACMAN VDDA voltage

        Returns VDDA and IDDA values from the built-in ADC as
        a tuple of mV and mA respectively

        '''
        if io_group is None:
            return dict([(io_group, self.get_vdda(io_group=io_group)) for io_group in self._io_group_table])
        mv = self._adc2mv(self.get_reg(self._vdda_adc_reg, io_group=io_group))
        ma = self._adc2ma(self.get_reg(self._idda_adc_reg, io_group=io_group))
        return mv, ma

    def set_vdda(self, vdda_dac=0xD5A3, io_group=None, settling_time=0.1):
        '''
        Sets PACMAN VDDA voltage

        If no ``vdda_dac`` value is specified, sets VDDA to default of ~1.8V

        Returns the resulting VDDA and IDDA values from the built-in ADC as
        a tuple of mV and mA respectively

        '''
        if io_group is None:
            return dict([(io_group, self.set_vdda(vdda_dac, io_group=io_group)) for io_group in self._io_group_table])
        self.set_reg(self._vdda_dac_reg, vdda_dac, io_group=io_group)
        if settling_time:
            time.sleep(settling_time)
        return self.get_vdda(io_group=io_group)

    def get_vplus(self, io_group=None):
        '''
        Gets PACMAN Vplus voltage

        Returns Vplus and Iplus values from the built-in ADC as
        a tuple of mV and mA respectively

        '''
        if io_group is None:
            return dict([(io_group, self.get_vplus(io_group=io_group)) for io_group in self._io_group_table])
        mv = self._adc2mv(self.get_reg(self._vplus_adc_reg, io_group=io_group))
        ma = self._adc2ma(self.get_reg(self._iplus_adc_reg, io_group=io_group))
        return mv, ma

    def enable_tile(self, tile_indices=None, io_group=None):
        '''
        Enables the specified pixel tile(s) (first tile is index=0, second
        tile is index=1, ...).

        Returns the value of the new tile enable mask

        '''
        if io_group is None:
            return dict([(io_group, self.enable_tile(tile_indices=tile_indices, io_group=io_group)) for io_group in self._io_group_table])
        if tile_indices is None:
            tile_indices = list(range(8))
        elif isinstance(tile_indices,int):
            tile_indices = [tile_indices]
        val = self.get_reg(self._base_ctrl_reg, io_group=io_group)
        for idx in tile_indices:
            val = val | (1 << idx)
        self.set_reg(self._base_ctrl_reg, val, io_group=io_group)
        return (self.get_reg(self._base_ctrl_reg, io_group=io_group) & 0xFF)

    def disable_tile(self, tile_indices=None, io_group=None):
        '''
        Disables the specified pixel tile(s) (first tile is index=0, second
        tile is index=1, ...).

        Returns the value of the new tile enable mask

        '''
        if io_group is None:
            return dict([(io_group, self.disable_tile(tile_indices=tile_indices, io_group=io_group)) for io_group in self._io_group_table])
        if tile_indices is None:
            tile_indices = list(range(8))
        elif isinstance(tile_indices,int):
            tile_indices = [tile_indices]
        val = self.get_reg(self._base_ctrl_reg, io_group=io_group)
        for idx in tile_indices:
            val = val & (0xFFFFFFFF & ~(1 << idx))
        self.set_reg(self._base_ctrl_reg, val, io_group=io_group)
        return (self.get_reg(self._base_ctrl_reg, io_group=io_group) & 0xFF)

    def set_uart_clock_ratio(self, channel, ratio, io_group=None):
        '''
        Sets PACMAN UART clock speed relative to the larpix master clock
        for the specified channel

        For a nominal 10MHz clock, a ratio value of 4 results in a 2.5MHz
        UART clock.

        Returns the value of the UART clock register that was set

        '''
        if io_group is None:
            return dict([(io_group, self.set_uart_clock_ratio(channel, ratio, io_group=io_group)) for io_group in self._io_group_table])
        reg = self._channel_size*channel + self._uart_clock_ratio_offset + self._channel_offset
        self.set_reg(reg, ratio, io_group=io_group)
        return self.get_reg(reg, io_group=io_group)

    def reset_larpix(self, length=256, io_group=None):
        '''
        Issues a reset of the specified length (in larpix MCLK cycles).

        If no ``length`` specified, issue a hard reset.

        Returns the value of the clock/reset control register after the reset

        '''
        if io_group is None:
            return dict([(io_group, self.reset_larpix(length, io_group=io_group)) for io_group in self._io_group_table])
        # set reset cycles
        self.set_reg(self._sw_reset_cycles_reg, length, io_group=io_group)
        # toggle reset bit
        clk_ctrl = self.get_reg(self._clk_ctrl_reg, io_group=io_group)
        self.set_reg(self._clk_ctrl_reg, clk_ctrl|4, io_group=io_group)
        self.set_reg(self._clk_ctrl_reg, clk_ctrl, io_group=io_group)
        return self.get_reg(self._clk_ctrl_reg, io_group=io_group)




'''
A module for the FakeIO class.

'''
from __future__ import print_function
from collections import deque

from larpix.io import IO
from larpix import TimestampPacket

class FakeIO(IO):
    '''
    An IO stand-in that sends output to stdout (i.e. print) and reads
    input from a data member that can be set in advance.

    The queue is implemented as a ``collections.deque`` object. Data can
    be queued up in advance through repeated calls to
    ``queue.append()``. The first element of the queue will be passed on to
    the ``Controller.read`` method each time it is called. This is a
    true queue, i.e. first-in, first-out.

    The format for an element of the queue is a tuple:
    ``([list_of_Packets], b'corresponding bytes')``.

    Although meaningless in terms of the internal implementation,
    ``FakeIO`` objects contain an internal state determining whether the
    object is currently "listening," and will raise a ``RuntimeError``
    if ``empty_queue`` is called when the object is not listening.

    '''
    def __init__(self):
        super(FakeIO, self).__init__()
        self.queue = deque()
        self.sent = []

    @classmethod
    def encode(cls, packets):
        '''
        Placeholder for packet encoding

        :returns: ``packets``
        '''
        return packets

    @classmethod
    def decode(cls, msgs):
        '''
        Placeholder for message decoding

        :returns: ``msgs``
        '''
        return msgs

    @staticmethod
    def add_timestamps(packets, positions, timestamps=0):
        '''
        Insert timestamp packets into a list of packets in the given
        positions.

        Convenience method for modifying lists of packets to add to the
        FakeIO queue. Modifies the list in-place.

        The positions are with respect to the indexes of the original
        list, so that the inserted element is just before the element
        that originally had that index. e.g.

        >>> add_timestamps([a, b, c, d], [1, 3])
        [a, TimestampPacket(...), b, c, TimestampPacket(...), d]

        If timestamps is a list, those timestamps will be used for the
        TimestampPackets. If it is an int, it will be taken as the
        starting time, and each subsequent packet will be incremented by
        1. A default starting time of 0 is assumed.

        '''
        npositions = len(positions)
        if isinstance(timestamps, int):
            timestamps = list(range(timestamps, timestamps+npositions))
        for position, timestamp in reversed(list(zip(positions, timestamps))):
            packets.insert(position, TimestampPacket(timestamp))

    def send(self, packets):
        '''
        Print the packets to stdout.

        '''
        for packet in packets:
            print(packet)
        self.sent.append(packets)

    def start_listening(self):
        '''
        Mock-up of beginning to listen for new packets.

        '''
        super(FakeIO, self).start_listening()

    def stop_listening(self):
        '''
        Mock-up of no longer listening for new packets.

        '''
        super(FakeIO, self).stop_listening()

    def empty_queue(self):
        '''
        Read and remove the next item from the internal queue and return
        it as if it were data that had just been read.

        '''
        if not self.is_listening:
            raise RuntimeError('Cannot empty queue when not'
                    ' listening')
        if len(self.queue):
            data = self.queue.popleft()
            return data
        return list(), b''




from bitarray import bitarray
import struct

from .. import bitarrayhelper as bah
from ..key import Key

class MessagePacket(object):
    '''
    A packet-like object which contains a string message and timestamp.

    :param message: a string message of length less than 64
    :param timestamp: the timestamp of the message

    '''
    size=72
    chip_key = None
    def __init__(self, message, timestamp):
        self.packet_type = 5
        self.message = message
        self.timestamp = timestamp

    def __str__(self):
        return '[ Message: %s | Timestamp: %d ]' % (self.message,
                self.timestamp)

    def __repr__(self):
        return 'MessagePacket(%s, %d)' % (repr(self.message),
                self.timestamp)

    def __eq__(self, other):
        return (self.message == other.message
                and self.timestamp == other.timestamp)

    def __ne__(self, other):
        return not (self == other)

    def export(self):
        return {
                'type_str': 'message',
                'type': self.packet_type,
                'message': self.message,
                'timestamp': self.timestamp,
                'bits': self.bits.to01(),
                }

    def from_dict(self, d):
        ''' Inverse of export - modify packet based on dict '''
        if 'type' in d and d['type'] != self.packet_type:
            raise ValueError('invalid packet type for MessagePacket')
        for key, value in d.items():
            if key == 'type':
                self.packet_type = value
            elif key == 'type_str':
                continue
            elif key == 'bits':
                self.bits = bitarray(value)
            else:
                setattr(self, key, value)

    @property
    def bits(self):
        b = bitarray()
        b.frombytes(self.bytes())
        return b

    @bits.setter
    def bits(self, value):
        value_bytes = value.tobytes()
        message_bytes = value_bytes[:64]
        timestamp_bytes = value_bytes[64:]
        self.message = message_bytes[:message_bytes.find(b'\x00')].decode()
        self.timestamp = struct.unpack('Q', timestamp_bytes)[0]

    def bytes(self):
        return (self.message.ljust(64, '\x00').encode()
                + struct.pack('Q', self.timestamp))




'''
Use the pytest framework to write tests for the v2b upgrade to larpix chip.

'''
from __future__ import print_function
import pytest
from larpix import (Chip, Packet_v1, Packet_v2, Packet, Key, Configuration, Configuration_v1, Configuration_v2, Configuration_v2b, Controller,
        PacketCollection, _Smart_List, TimestampPacket, MessagePacket)
from larpix.io import FakeIO
#from bitstring import BitArray
from bitarray import bitarray
import larpix.bitarrayhelper as bah
import json
import os

import random



@pytest.fixture
def timestamp_packet():
    return TimestampPacket(123456789)

@pytest.fixture
def message_packet():
    return MessagePacket('test',123456789)

def list_of_packets_str(packets):
    return '\n'.join(map(str, packets)) + '\n'

def test_chip_str():
    key = '1-1-1'
    chip = Chip(key, version='2b')
    result = str(chip)
    expected = 'Chip (key: 1-1-1, version: 2b)'
    assert result == expected

def test_chip_get_configuration_packets(chip2b):
    packet_type = Packet.CONFIG_WRITE_PACKET
    packets = chip2b.get_configuration_write_packets()
    # test a sampling of the configuration packets
    packet = packets[5]
    assert packet.packet_type == packet_type
    assert packet.chip_id == chip2b.chip_id
    assert packet.register_address == 5
    assert packet.register_data == 16

    packet = packets[40]
    assert packet.packet_type == packet_type
    assert packet.chip_id == chip2b.chip_id
    assert packet.register_address == 40
    assert packet.register_data == 16

def test_chip_sync_configuration(chip2b):
    packet_type = Packet.CONFIG_READ_PACKET
    packets = chip2b.get_configuration_read_packets()
    chip2b.reads.append(PacketCollection(packets))
    chip2b.sync_configuration()
    result = chip2b.config.all_data()
    expected = [bitarray([0]*8)] * Configuration_v2b.num_registers
    assert result == expected

def test_chip_sync_configuration_slice(chip2b):
    packets = chip2b.get_configuration_read_packets()
    chip2b.reads.append(PacketCollection(packets[:10]))
    chip2b.reads.append(PacketCollection(packets[10:]))
    chip2b.sync_configuration(index=slice(None, None, None))
    result = chip2b.config.all_data()
    expected = [bitarray([0]*8)] * Configuration_v2b.num_registers
    assert result == expected

def test_chip_export_reads(chip2b):
    packet = Packet()
    packet.chip_key = chip2b.chip_key
    packet.packet_type = Packet.CONFIG_WRITE_PACKET
    packet.chip_id = chip2b.chip_id
    packet.register_address = 10
    packet.register_data = 20
    packet.assign_parity()
    chip2b.reads.append(packet)
    result = chip2b.export_reads()
    expected = {
            'chip_key': chip2b.chip_key,
            'chip_id': chip2b.chip_id,
            'packets': [
                packet.export()
                ]
            }
    assert result == expected
    assert chip2b.new_reads_index == 1

def test_configuration_write(tmpdir):
    c = Configuration_v2b()
    f = str(tmpdir.join('test_config.json'))
    c.write(f)
    with open(f, 'r') as output:
        result = json.load(output)
    expected = c.to_dict()
    assert result['register_values'] == expected

def test_configuration_write_errors(tmpdir):
    c = Configuration_v2b()
    f = tmpdir.join('test_config.json')
    f.write("Test data.....")
    with pytest.raises(IOError):
        c.write(str(f))
        pytest.fail('Should fail: force fails')

def test_controller_read_configuration(capfd, chip2b):
    controller = Controller()
    controller.io = FakeIO()
    controller.chips[chip2b.chip_key] = chip2b
    conf_data = chip2b.get_configuration_packets(Packet.CONFIG_READ_PACKET)
    sent_expected = list_of_packets_str(conf_data)
    controller.io.queue.append((conf_data,b'hi'))
    received_expected = PacketCollection(conf_data, b'hi', read_id=0,
            message='configuration read')
    controller.read_configuration(chip2b.chip_key, timeout=0.01)
    received_result = controller.reads[-1]
    sent_result, err = capfd.readouterr()
    assert sent_result == sent_expected
    assert received_result == received_expected

def test_conf_v2b():
    c = Configuration_v2b()
    endian = 'little'

    # test simple register
    c.threshold_global = 255
    assert c.threshold_global == 255
    assert c.threshold_global_data == [(64, bah.fromuint(255,8, endian=endian))]

    c.threshold_global_data = (64, bah.fromuint(253,8, endian=endian))
    assert c.threshold_global == 253
    assert c.threshold_global_data == [(64, bah.fromuint(253,8, endian=endian))]

    c.threshold_global_data = bah.fromuint(252,8, endian=endian)
    assert c.threshold_global == 252
    assert c.threshold_global_data == [(64, bah.fromuint(252,8, endian=endian))]

    # test list register
    c.pixel_trim_dac = [0]*64
    assert c.pixel_trim_dac == [0]*64
    assert c.pixel_trim_dac_data == [(i, bah.fromuint(0,8, endian=endian)) for i in range(64)]

    c.pixel_trim_dac[1] = 1
    assert c.pixel_trim_dac[1] == 1
    assert c.pixel_trim_dac_data[1] == (1, bah.fromuint(1,8, endian=endian))
    assert c.pixel_trim_dac_data[0] == (0, bah.fromuint(0,8, endian=endian))

    c.pixel_trim_dac_data = (1, bah.fromuint(2,8, endian=endian))
    assert c.pixel_trim_dac[1] == 2
    assert c.pixel_trim_dac_data[1] == (1, bah.fromuint(2,8, endian=endian))
    assert c.pixel_trim_dac_data[0] == (0, bah.fromuint(0,8, endian=endian))

    bits = bitarray()
    for i in range(64):
        bits += bah.fromuint(31,8, endian=endian)
    c.pixel_trim_dac_data = bits
    assert c.pixel_trim_dac[1] == 31
    assert c.pixel_trim_dac[0] == 31
    assert c.pixel_trim_dac_data[1] == (1, bah.fromuint(31,8, endian=endian))
    assert c.pixel_trim_dac_data[0] == (0, bah.fromuint(31,8, endian=endian))

    # test compound register
    c.csa_gain = 1
    c.csa_bypass_enable = 1
    c.bypass_caps_en = 1
    reg_data = [(65, bitarray('11100000'))]
    assert c.csa_gain == 1
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 1
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 1
    assert c.bypass_caps_en_data == reg_data

    c.csa_bypass_enable = 0
    reg_data = [(65, bitarray('10100000'))]
    assert c.csa_gain == 1
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 0
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 1
    assert c.bypass_caps_en_data == reg_data

    reg_data = [(65, bitarray('01000000'))]
    c.csa_bypass_enable_data = reg_data[0]
    assert c.csa_gain == 0
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 1
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 0
    assert c.bypass_caps_en_data == reg_data

    reg_data = [(65, bitarray('11000000'))]
    c.csa_gain_data = bitarray('1')
    assert c.csa_gain == 1
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 1
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 0
    assert c.bypass_caps_en_data == reg_data

    # test list register (that covers <1 register)
    c.current_monitor_bank0 = [0]*4
    assert c.current_monitor_bank0 == [0]*4
    assert c.current_monitor_bank0_data == [(109, bah.fromuint(0,8, endian=endian))]

    c.current_monitor_bank0[1] = 1
    assert c.current_monitor_bank0[1] == 1
    assert c.current_monitor_bank0[0] == 0
    assert c.current_monitor_bank0_data == [(109, bitarray('01000000'))]

    c.current_monitor_bank0_data = (109, bitarray('00100000'))
    assert c.current_monitor_bank0[1] == 0
    assert c.current_monitor_bank0[2] == 1
    assert c.current_monitor_bank0_data == [(109, bitarray('00100000'))]

    bits = bitarray('0100')
    c.current_monitor_bank0_data = bits
    assert c.current_monitor_bank0[1] == 1
    assert c.current_monitor_bank0[0] == 0
    assert c.current_monitor_bank0_data == [(109, bitarray('01000000'))]

    # test long register
    c.periodic_trigger_cycles = 2**32-1
    assert c.periodic_trigger_cycles == 2**32-1
    assert c.periodic_trigger_cycles_data == [(i, bah.fromuint(255,8, endian=endian)) for i in range(166,170)]

    c.periodic_trigger_cycles_data = (166, bah.fromuint(254,8, endian=endian))
    assert c.periodic_trigger_cycles == 2**32-2
    assert c.periodic_trigger_cycles_data[0] == (166, bah.fromuint(254, 8, endian=endian))

    c.periodic_trigger_cycles_data = bah.fromuint(1,32, endian=endian)
    assert c.periodic_trigger_cycles == 1
    assert c.periodic_trigger_cycles_data[0] == (166, bah.fromuint(1,8, endian=endian))

    c.enable_piso_downstream[1] = 1
    assert c.enable_piso_downstream[1] == 1
    assert c.enable_piso_downstream[0] == 0
    assert c.enable_piso_downstream_data == [(125, bitarray('01000000'))]

    c.enable_piso_downstream_data = (125, bitarray('00100000'))
    assert c.enable_piso_downstream[1] == 0
    assert c.enable_piso_downstream[2] == 1
    assert c.enable_piso_downstream_data == [(125, bitarray('00100000'))]

    bits = bitarray('0100')
    c.enable_piso_downstream_data = bits
    assert c.enable_piso_downstream[1] == 1
    assert c.enable_piso_downstream[0] == 0
    assert c.enable_piso_downstream_data == [(125, bitarray('01000000'))]

    c.i_rx0 = 8
    c.i_rx1 = 1
    assert c.i_rx0_data == [(243, bitarray('00011000'))]
    assert c.i_rx1_data == [(243, bitarray('00011000'))]

    c.i_rx0 = 15
    assert c.i_rx0_data == [(243, bitarray('11111000'))]
    assert c.i_rx1_data == [(243, bitarray('11111000'))]

    c.test_mode_uart0 = 2
    c.test_mode_uart1 = 0
    c.test_mode_uart2 = 0
    c.test_mode_uart3 = 0

    assert c.test_mode_uart0_data == [(127, bitarray('01000000'))]

def test_compare_v2b():
    c = Configuration_v2b()
    other = Configuration_v2b()
    assert c.compare(other) == {}

    c.pixel_trim_dac[10] = 25
    c.pixel_trim_dac[12] = 26
    assert c.compare(other) == {'pixel_trim_dac': [
        ({'index': 10, 'value': 25}, {'index': 10, 'value': 16}),
        ({'index': 12, 'value': 26}, {'index': 12, 'value': 16}),
        ]}

    c = Configuration_v2b()
    c.threshold_global = 121
    assert c.compare(other) == {'threshold_global': (121, 255)}

def test_load_inheritance_v2b(tmpdir):
    filename = 'test.json'
    c = Configuration_v2b()
    with open(str(tmpdir.join(filename)),'w') as of:
        config = {
            "_config_type": "chip",
            "_include": ["chip/default_v2b.json"],
            "class": "Configuration_v2b",
            "register_values": {
                "pixel_trim_dac": [0, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16],
                "threshold_global": 0
            }
        }
        json.dump(config,of)
    c.load(str(tmpdir.join(filename)))
    assert c.pixel_trim_dac[0] == 0
    assert c.threshold_global == 0
    assert c.csa_gain == 0

def test_get_nondefault_registers_v2b():
    c = Configuration_v2b()
    c.pixel_trim_dac[10] = 25
    c.threshold_global = 121
    c.csa_gain = 1
    c.csa_enable[35] = 0
    assert c.get_nondefault_registers() == {
            'pixel_trim_dac': [({'index': 10, 'value': 25},
                {'index': 10, 'value': 16})],
            'threshold_global': (121, 255),
            'csa_gain': (1, 0),
            'csa_enable': [({'index': 35, 'value': 0}, {'index': 35,
                'value': 1})],
            }

def test_v2b_conf_all_data():
    c = Configuration_v2b()
    config_data = c.all_data()

    assert len(config_data) == c.num_registers, \
        f'config data length incorrect (should be {c.num_registers}, is {len(config_data)})'

    for reg_name in c.register_map.keys():
        reg_data = getattr(c, f'{reg_name}_data')
        for reg_addr, bits in reg_data:
            assert bits == config_data[reg_addr], \
                f'register {reg_addr} mismatch (should be {bits}, is {config_data[reg_addr]})'


def test_v2b_conf_some_data():
    c = Configuration_v2b()

    req_addrs = range(c.num_registers)
    addrs, bits = c.some_data(req_addrs)
    config_data = c.all_data()

    # check all addresses
    assert len(addrs) == len(req_addrs), \
        f'config data length incorrect (should be {len(req_addrs)}, is {len(addrs)})'
    assert set(addrs) == set(req_addrs), \
        f'config data values incorrect (should be {set(req_addrs)}, is {set(addrs)})'

    for addr, addr_bits in zip(addrs, bits):
        assert addr_bits == config_data[addr], \
            f'register {addr} mismatch (should be {config_data[addr]}, is {addr_bits})'

     # check partial addresses
    req_addrs = random.sample(range(c.num_registers), c.num_registers//2)
    addrs, bits = c.some_data(req_addrs)

    # check all addresses
    assert len(addrs) == len(req_addrs), \
        f'config data length incorrect (should be {len(req_addrs)}, is {len(addrs)})'
    assert set(addrs) == set(req_addrs), \
        f'config data values incorrect (should be {set(req_addrs)}, is {set(addrs)})'

    for addr, addr_bits in zip(addrs, bits):
        assert addr_bits == config_data[addr], \
            f'register {addr} mismatch (should be {config_data[addr]}, is {addr_bits})'




from .configuration import *
from .configuration_v1 import *
from .configuration_v2_base import *
from .configuration_v2 import *
from .configuration_lightpix_v1 import *
from .configuration_v2b import *

Configuration = Configuration_v2



'''
This module contains controller configuration files, used to specify the physical
asics present and their network structure

V2 Configuration
----------------

The v2 configuration file is a standard JSON file structured as follows:

.. parsed-literal::
    {
        "_config_type": "controller",
        "_include": [<optional list of files to inherit from>],
        "name": <string identifier for configuration ID, e.g. "pcb-3", or "acube-2x2">,
        "layout": <string identifier for layout version, e.g. "1.2.0">,
        "asic_version": 2,
        "network": {
            "miso_uart_map": [<uart channel for position in array, required>, ...],
            "mosi_uart_map": [<uart channel for position in array, required>, ...],
            "usds_link_map": [<position of chip relative to neighbor, required>, ...],
            "<io_group>": {
                "miso_uart_map": [<uart channels for io_group, optional>, ...],
                ...
                "<io_channel>": {
                    "miso_uart_map": [<uart channels for io_group, optional>, ...],
                    ...
                    "nodes": [
                        {
                            "chip_id": <chip id of first chip to load into network>,
                            "miso_us": [<chip_ids of upstream chips, null if no upstream chip, optional>],
                            "miso_ds": [<chip_ids of downstream chips, optional>],
                            "mosi": [<chip_ids of chips to listen to, optional>],
                            "root": <bool indicating if this is a root node, optional>,
                            "miso_uart_map": [<uart channels for chip, optional>, ...],
                            ...
                        },
                        ...
                    ]
                },
                ...
            },
            ...
        }
    }

The structure allows for a lot of flexibility in the network configuration, while
maintaining a relatively terse format for simple configurations. Generally, a
Hydra network can be described using three directed graphs which we call
miso upstream (``miso_us``), miso downstream (``miso_ds``), and mosi (``mosi``). The ``miso_us`` graph
represents the flow of outgoing upstream packets (i.e. the direction of configuration
write packets). The ``miso_ds`` graph represents the flow of outgoing downstream packet
(i.e. the direction of configuration read and data packets). The ``mosi`` graph
represents the flow of incoming packets of either type. The overall LArPix
network is then defined with a collection of Hydra networks, each associated
with a specific io_group and io_channel. The ``io_group``, ``io_channel`` pair must
be unique for each Hydra network.

For the configuration file, the basic requirement for each hydra network is that for each
node in the network, you must specify a chip id. The chip id must be an integer
between 2 and 254 for an ASIC, but other systems (i.e. an FPGA that can interpret
the LArPix UART protocol) can be specified by a string. Each chip id must be
unique within each io_group, io_channel sub-network.

To specify links within the 3 directed graphs (``miso_us``, ``miso_ds``, and ``mosi``), a
4-item array is used. Each 4-item array indicates the chip id of the neighboring
nodes that should be connected to the given chip. So for example::

    {
        "chip_id": 23,
        "miso_us": [24,null,null,null],
        "miso_ds": [null,22,null,null],
        "mosi":    [24,22,null,null]
    }

will create a ``miso_us`` graph where data flows from chip 23 to chip 24, a ``miso_ds``
graph where data flows from chip 23 to chip 22, and a ``mosi`` graph where data
flows from chip 24 to 23 and chip 22 to 23 (note the opposite convention for the
mosi graph).

The position of the chip id within the 4-item array indicates the relative physical
location of the neighboring chip. A suggested ordering might be
``[<top>,<left>,<down>,<right>]``, thus the previous example indicates
that the current chip should be linked to chip 24 which is above the current chip.
The definition of ``<top>``, ``<left>``, ``<down>``, and ``<right>`` is fixed by which of
the four LArPix UARTs is used to communicate to a chip in that position. This
is declared with the ``"miso_uart_map"`` and ``"mosi_uart_map"`` definitions.
As an example, the upper left corner of the LArPix ASIC is UART 0 and the upper
right corner of the chip is UART 3, thus we can define
``<top>`` as the first position in each array by declaring
``"miso_uart_map": [3,X,X,X]`` and ``"mosi_uart_map": [0,X,X,X]`` which will read
packets in on UART 0 and send packets out on UART 3. It doesn't matter where
in the array ``<top>`` is defined, as long as you have a consistent convention between
the ``miso_uart_map``, ``mosi_uart_map``, and each of your network node's (``miso_us``, ``miso_ds``,
and ``mosi``).

It would become very tedious and difficult to debug if for every node you had
to specify all of the links on all three of the graphs, so the configuration
parser is smarter than that. Most of the time, the networks that you'll want to
specify have the same upstream and downstream connections (albeit with the
directions flipped), so if the ``"miso_ds"`` field is not explicitly defined, it
will be inferred based on the ``miso_us`` network that was created. Likewise, usually
the ``mosi`` connections should be the same as the upstream and downstream connections.
Thus, if this field is not set explicitly it will be inferred automatically based
on the ``miso_us`` graph and the ``miso_ds`` graph.

However, in order for the auto-parsing methods to work, they need some information
about how the chips are oriented with respect to each other. This is defined by
the ``"usds_link_map"``. This field is also a 4-item array, however the value now
represents the position of the given chip from the perspective of the neighbor.

So, following the example from above, we can define two directions ``<top>`` and
``<bottom>`` with ``"miso_uart_map": [3,X,1,X]`` and ``"mosi_uart_map": [0,X,2,X]``.
If we have two chips (let's say 22 and 23 with chip 23 above chip 22), to
determine the ``"usds_link_map"`` we have to consider both where chip 23 is from the
perspective of chip 22 (``<top>``, or index 0 from our definition) and where chip 22
is from the perspective of chip 23 (``<bottom>``, or index 2 from out definition).
This means that the ``"usds_link_map"`` for chip 22 should be ``[2,X,X,X]``, or in
other words, the chip (23) that is in the direction of ``<top>`` (index 0) from
the current chip (22) sees the current chip (22) in the direction of ``<down>``
(index 2). Likewise the ``"usds_link_map"`` for chip 23 should be ``[X,X,0,X]``.
Now as long as you maintain the same definitions of ``<top>`` and ``<bottom>`` for
all of the chips in the network, you can set ``"usds_link_map"`` at the top level
to be ``[2,X,0,X]`` you don't need to worry about figuring this out for each
node in the network... phew.

The final necessary component of the configuration is to declare some nodes
as ``root`` nodes. These are special nodes which are used to indicate the first
node to be configured during the initialization of the network. Typically, the
placeholder node representing the external system is set as the root node.


Putting all of this together, a simple example of a complete network
configuration is provided below:

.. parsed-literal::
    "network": {
        "1": {
            "2": {
                "nodes": [
                    {
                        "chip_id": "ext",
                        "miso_us": [null,null,2,null],
                        "root": true
                    }
                    {
                        "chip_id": 2,
                        "miso_us": [null,3,null,null]
                    },
                    {
                        "chip_id": 3
                    }
                ]
            }
        }
        "miso_uart_map": [3,0,1,2],
        "mosi_uart_map": [0,1,2,3],
        "usds_link_map": [2,3,0,1]
    }

In this example, we define a single 2-chip Hydra network with ``io_group=1``
and ``io_channel=2``. The ``"nodes"`` array is filled with ``dict``s representing each
node in the network. Globally, we've defined our "directions" with the
``miso_uart_map`` and ``mosi_uart_map``, and we've set up the ``usds_link_map`` so
that the auto-parsing algorithm can work. Within the ``"nodes"`` array, we specify
three nodes (``ext``, ``2``, ``3``); two of which represent chips (``2``, ``3``), and one of
which represents the external FPGA used to communicate with the network (``ext``).
The external FPGA is set as the ``root`` node so that the initialization of the
network starts from there. We've declared ``miso_us`` links for the ``ext`` and
``2`` nodes, but because we have not specified any ``miso_ds`` or ``mosi`` links, the
auto-parsing algorithm with generate the following graphs::

    miso_us: 'ext'  -> 2  -> 3
    miso_ds: 'ext' <-  2 <-  3
    mosi:    'ext' <-> 2 <-> 3

with the active uarts::

    miso_us: 1    -> 0    -> X
    miso_ds: X   <-  3   <-  2
    mosi:    2,0 <-> 1,3 <-> 3

Or, assuming you are using our definition above for ``<top>``, ``<bottom>``, etc.::

       ext
        v
    3 < 2

The top-level ``"_include"`` field can be used to specify files to inherit from.
All fields of the configuration down to the ``"nodes"`` array level can be inherited.
A standard use for this would be to specify each hydra network for a given channel
independently, and inherit from all of the files in a single configuration
file. E.g.::

    {
        "_config_type": "controller",
        "_include": [
            "network-1-1.json",
            "network-1-2.json",
            "network-1-3.json",
            "network-1-4.json",
            "network-2-1.json",
            "network-2-2.json",
            "network-2-3.json",
            "network-2-4.json"
            ],
        "name": "multi-io-group-network"
    }


LightPix Configuration
----------------------

LightPix uses the same interface as LArPix v2 but with::

    "asic_version": "lightpix-v1.0"


V1 Configuration
----------------

The v1 configuration file is a standard JSON file structured as follows:

.. parsed-literal::
    {
        "name": <string identifier for PCB ID, e.g. "pcb-3">,
        "layout": <string identifier for layout version, e.g. "1.2.0">,
        "asic_version": 1,
        "chip_list": [<A list of chip keys, one for each chip, in daisy-chain order>]
    }

All fields are necessary.

See the larpix-geometry documentation
<https://github.com/samkohn/larpix-geometry> for a listing of all the
layout versions.
'''
pass



from bitarray import bitarray
import os
import errno
import functools
from collections import OrderedDict

from .. import bitarrayhelper as bah
from .. import configs
from . import BaseConfiguration_v2, _Smart_List
from . import configuration_v2_base as v2_base

class Configuration_v2(BaseConfiguration_v2):
    '''
    Represents the desired configuration state of a LArPix v2 chip.

    Each register name is available as its own attribute for inspecting and
    setting the value of the corresponding register.

    Certain configuration values are set channel-by-channel. These are
    represented by a list of values. For example:

        >>> conf.pixel_trim_dac[2:5]
        [16, 16, 16]
        >>> conf.channel_mask[20] = 1
        >>> conf.external_trigger_mask = [0] * 64

    Additionally, other configuration values take up more than or less
    than one complete register. These are still set by referencing the
    appropriate name. For example, ``cross_trigger_mode`` shares a
    register with a few other values, and adjusting the value of the
    ``cross_trigger_mode`` attribute will leave the other values
    unchanged.

    Each register name can cover more than one 'physical' register depending on
    the size of the data it holds. You can see which physical registers a
    given register name corresponds to by using the `register_map` attribute, e.g.::

        >>> conf.register_map['digital_threshold']  # 64 registers, 1 per channel
        range(173, 237)
        >>> conf.register_map['enable_dynamic_reset']  # Register 170
        range(170, 171)
        >>> conf.register_map['enable_min_delta_adc']  # Shares register 170
        range(170, 171)

    '''
    asic_version = 2
    default_configuration_file = 'chip/default_v2.json'
    num_registers = 237
    num_bits = 1896
    num_channels = 64
    # Additional class properties regarding configuration registers are set at the end of the file.

    def __init__(self):
        # Note: properties, getters and setters are constructed after this class definition at the bottom of the file.
        super(BaseConfiguration_v2, self).__init__()
        return

## Set up property info
#
_property_configuration = OrderedDict([
        ('pixel_trim_dac',
            (v2_base._list_property, (int, 0, 31, Configuration_v2.num_channels, 8), (0,512))),
        ('threshold_global',
            (v2_base._basic_property, (int, 0, 255), (512, 520))),
        ('csa_gain',
            (v2_base._compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (520,521))),
        ('csa_bypass_enable',
            (v2_base._compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (521,522))),
        ('bypass_caps_en',
            (v2_base._compound_property, (['csa_gain', 'csa_bypass_enable','bypass_caps_en'], (int,bool), 0, 1), (522,523))),
        ('csa_enable',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (528, 592))),
        ('ibias_tdac',
            (v2_base._basic_property, (int, 0, 15), (592, 596))),
        ('ibias_comp',
            (v2_base._basic_property, (int, 0, 15), (600, 604))),
        ('ibias_buffer',
            (v2_base._basic_property, (int, 0, 15), (608, 612))),
        ('ibias_csa',
            (v2_base._basic_property, (int, 0, 15), (616, 620))),
        ('ibias_vref_buffer',
            (v2_base._basic_property, (int, 0, 15), (624, 628))),
        ('ibias_vcm_buffer',
            (v2_base._basic_property, (int, 0, 15), (632, 636))),
        ('ibias_tpulse',
            (v2_base._basic_property, (int, 0, 15), (640, 644))),
        ('ref_current_trim',
            (v2_base._compound_property, (['ref_current_trim','override_ref','ref_kickstart'], int, 0, 31), (648, 653))),
        ('override_ref',
            (v2_base._compound_property, (['ref_current_trim','override_ref','ref_kickstart'], (int,bool), 0, 1), (653, 654))),
        ('ref_kickstart',
            (v2_base._compound_property, (['ref_current_trim','override_ref','ref_kickstart'], (int,bool), 0, 1), (654, 655))),
        ('vref_dac',
            (v2_base._basic_property, (int, 0, 255), (656, 664))),
        ('vcm_dac',
            (v2_base._basic_property, (int, 0, 255), (664,672))),
        ('csa_bypass_select',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (672,736))),
        ('csa_monitor_select',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (736,800))),
        ('csa_testpulse_enable',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (800,864))),
        ('csa_testpulse_dac',
            (v2_base._basic_property, (int, 0, 255), (864,872))),
        ('current_monitor_bank0',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (872,876))),
        ('current_monitor_bank1',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (880,884))),
        ('current_monitor_bank2',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (888,892))),
        ('current_monitor_bank3',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (896,900))),
        ('voltage_monitor_bank0',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (904,907))),
        ('voltage_monitor_bank1',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (912,915))),
        ('voltage_monitor_bank2',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (920,923))),
        ('voltage_monitor_bank3',
            (v2_base._list_property, ((int,bool), 0, 1, 3, 1), (928,931))),
        ('voltage_monitor_refgen',
            (v2_base._list_property, ((int,bool), 0, 1, 8, 1), (936,944))),
        ('digital_monitor_enable',
            (v2_base._compound_property, (['digital_monitor_enable','digital_monitor_select'], (int,bool), 0, 1), (944,945))),
        ('digital_monitor_select',
            (v2_base._compound_property, (['digital_monitor_enable','digital_monitor_select'], (int,bool), 0, 10), (945,949))),
        ('digital_monitor_chan',
            (v2_base._basic_property, (int, 0, 63), (952,958))),
        ('slope_control0',
            (v2_base._compound_property, (['slope_control0', 'slope_control1'], int, 0, 15), (960,964))),
        ('slope_control1',
            (v2_base._compound_property, (['slope_control0', 'slope_control1'], int, 0, 15), (964,968))),
        ('slope_control2',
            (v2_base._compound_property, (['slope_control2', 'slope_control3'], int, 0, 15), (968,972))),
        ('slope_control3',
            (v2_base._compound_property, (['slope_control2', 'slope_control3'], int, 0, 15), (972,976))),
        ('chip_id',
            (v2_base._basic_property, (int, 0, 255), (976,984))),
        ('load_config_defaults',
            (v2_base._compound_property, (['load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl'], (int,bool), 0, 1), (985,986))),
        ('enable_fifo_diagnostics',
            (v2_base._compound_property, (['load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl'], (int,bool), 0, 1), (986,987))),
        ('clk_ctrl',
            (v2_base._compound_property, (['load_config_defaults', 'enable_fifo_diagnostics', 'clk_ctrl'], (int), 0, 2), (987,989))),
        ('enable_miso_upstream',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (992,996))),
        ('enable_miso_downstream',
            (v2_base._compound_list_property, (['enable_miso_downstream', 'enable_miso_differential'], (int,bool), 0, 1, 4, 1), (1000,1004))),
        ('enable_miso_differential',
            (v2_base._compound_list_property, (['enable_miso_downstream', 'enable_miso_differential'], (int,bool), 0, 1, 4, 1), (1004,1008))),
        ('enable_mosi',
            (v2_base._list_property, ((int,bool), 0, 1, 4, 1), (1008,1012))),
        ('test_mode_uart0',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1016,1018))),
        ('test_mode_uart1',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1018,1020))),
        ('test_mode_uart2',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1020,1022))),
        ('test_mode_uart3',
            (v2_base._compound_property, (['test_mode_uart0', 'test_mode_uart1', 'test_mode_uart2', 'test_mode_uart3'], int, 0, 4), (1022,1024))),
        ('enable_cross_trigger',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1024,1025))),
        ('enable_periodic_reset',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1025,1026))),
        ('enable_rolling_periodic_reset',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1026,1027))),
        ('enable_periodic_trigger',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1027,1028))),
        ('enable_rolling_periodic_trigger',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1028,1029))),
        ('enable_periodic_trigger_veto',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1029,1030))),
        ('enable_hit_veto',
            (v2_base._compound_property, (['enable_cross_trigger', 'enable_periodic_reset', 'enable_rolling_periodic_reset', 'enable_periodic_trigger', 'enable_rolling_periodic_trigger', 'enable_periodic_trigger_veto', 'enable_hit_veto'], (int,bool), 0, 1), (1030,1031))),
        ('adc_hold_delay',
            (v2_base._basic_property, (int, 0, 15), (1032,1036))),
        ('adc_burst_length',
            (v2_base._basic_property, (int, 0, 255), (1040,1048))),
        ('channel_mask',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1048,1112))),
        ('external_trigger_mask',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1112,1176))),
        ('cross_trigger_mask',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1176,1240))),
        ('periodic_trigger_mask',
            (v2_base._list_property, ((int,bool), 0, 1, Configuration_v2.num_channels, 1), (1240,1304))),
        ('periodic_reset_cycles',
            (v2_base._basic_property, (int, 0, 2**24-1), (1304,1328))),
        ('periodic_trigger_cycles',
            (v2_base._basic_property, (int, 0, 2**32-1), (1328,1360))),
        ('enable_dynamic_reset',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1360,1361))),
        ('enable_min_delta_adc',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1361,1362))),
        ('threshold_polarity',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1362,1363))),
        ('reset_length',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int), 0, 7), (1363,1366))),
        ('mark_first_packet',
            (v2_base._compound_property, (['enable_dynamic_reset', 'enable_min_delta_adc', 'threshold_polarity', 'reset_length', 'mark_first_packet'], (int,bool), 0, 1), (1366,1367))),
        ('reset_threshold',
            (v2_base._basic_property, (int, 0, 255), (1368,1376))),
        ('min_delta_adc',
            (v2_base._basic_property, (int, 0, 255), (1376,1384))),
        ('digital_threshold',
            (v2_base._list_property, (int, 0, 255, Configuration_v2.num_channels, 8), (1384,1896))),
    ])

# GENERATE THE PROPERTIES!
Configuration_v2.bit_map = OrderedDict()
Configuration_v2.register_map = OrderedDict()
Configuration_v2.register_names = []


v2_base._generate_properties(Configuration_v2, _property_configuration, verbose=False)






import pytest
import os
import subprocess
import sys
import os
_dir_ = os.path.dirname(os.path.abspath(__file__))

from larpix import TimestampPacket, Packet_v2, SyncPacket, TriggerPacket
import larpix.format.rawhdf5format as r_h5_fmt
import larpix.format.hdf5format as p_h5_fmt
import larpix.format.pacman_msg_format as p_msg_fmt

@pytest.fixture
def test_packets():
    pkts = []
    for _ in range(100):
        pkts += [[]]
        pkts[-1] += [TimestampPacket(0)]
        pkts[-1] += [Packet_v2(b'\x00\x00\x00\x00\x00\x00\x00\x01')] * 10
        pkts[-1] += [TriggerPacket(b'\x00', 0, 0)]
        pkts[-1] += [SyncPacket(b'\x00', 0, 0, 0)]
    return pkts

@pytest.fixture
def raw_hdf5_tmpfile(tmpdir, test_packets):
    msgs = [p_msg_fmt.format([pkts]) for pkts in test_packets]
    io_groups = [0 for _ in test_packets]

    test_filename = os.path.join(tmpdir,'raw_test.h5')
    r_h5_fmt.to_rawfile(
        test_filename, msgs=msgs, msg_headers={'io_groups': io_groups}
        )
    return test_filename

@pytest.fixture
def packet_hdf5_tmpfile(tmpdir, test_packets):
    test_filename = os.path.join(tmpdir,'datalog_test.h5')
    p_h5_fmt.to_file(
        test_filename,
        packet_list=[p for pkts in test_packets for p in pkts]
        )
    return test_filename

def test_convert_rawhdf5_to_hdf5(tmpdir, raw_hdf5_tmpfile):
    out_filename = os.path.join(tmpdir, 'datalog_convert_test.h5')
    proc = subprocess.run(
        ['python', os.path.join(_dir_,'../scripts/convert_rawhdf5_to_hdf5.py'), '-i', raw_hdf5_tmpfile, '-o', out_filename, '--block_size', '10'],
        check=True
        )

    # test read from file
    new_packets = p_h5_fmt.from_file(out_filename)['packets']
    orig_packets = [p_msg_fmt.parse(msg) for msg in r_h5_fmt.from_rawfile(raw_hdf5_tmpfile)['msgs']]
    assert new_packets == [p for pkts in orig_packets for p in pkts]

def test_packet_hdf5_tool(tmpdir, packet_hdf5_tmpfile, test_packets):
    out_filename = os.path.join(tmpdir, 'datalog_tool_test.h5')

    # test merge
    proc = subprocess.run(
        ['python', os.path.join(_dir_,'../scripts/packet_hdf5_tool.py'), '--merge', '-i', packet_hdf5_tmpfile, packet_hdf5_tmpfile, '-o', out_filename, '--block_size', '10'],
        check=True
        )

    # test read from file
    orig_packets = p_h5_fmt.from_file(packet_hdf5_tmpfile)['packets']
    orig_packets += p_h5_fmt.from_file(packet_hdf5_tmpfile)['packets']
    new_packets = p_h5_fmt.from_file(out_filename)['packets']
    assert len(orig_packets) == len(new_packets)
    assert orig_packets == new_packets

def test_raw_hdf5_tool(tmpdir, raw_hdf5_tmpfile, test_packets):
    out_filename = os.path.join(tmpdir, 'raw_tool_test.h5')

    # test merge
    proc = subprocess.run(
        ['python', os.path.join(_dir_,'../scripts/raw_hdf5_tool.py'), '--merge', '-i', raw_hdf5_tmpfile, raw_hdf5_tmpfile, '-o', out_filename, '--block_size', '10'],
        check=True
        )

    # test read data
    assert r_h5_fmt.from_rawfile(raw_hdf5_tmpfile)['msgs'] + r_h5_fmt.from_rawfile(raw_hdf5_tmpfile)['msgs'] == r_h5_fmt.from_rawfile(out_filename)['msgs']

    # test merge
    proc = subprocess.run(
        ['python', os.path.join(_dir_,'../scripts/raw_hdf5_tool.py'), '--split', '-i', out_filename, '-o', tmpdir, '--max_length', '0', '--block_size', '10'],
        check=True
        )

    # test read data
    assert r_h5_fmt.from_rawfile(raw_hdf5_tmpfile)['msgs'] + r_h5_fmt.from_rawfile(raw_hdf5_tmpfile)['msgs'] == r_h5_fmt.from_rawfile(out_filename)['msgs']




from __future__ import print_function

import pytest
import h5py
import numpy as np

from larpix.format.rawhdf5format import (to_rawfile, from_rawfile, len_rawfile)

@pytest.fixture
def tmpfile(tmpdir):
    return str(tmpdir.join('test_raw.h5'))

@pytest.fixture
def testdata():
    test_msgs = [
        b'\x00'*8+b'\x01'*16,
        b'\x01'*8+b'\x02'*16,
        b'\x02'*8+b'\x03'*16
        ]
    test_io_groups = [
        0,
        1,
        2
    ]
    return test_io_groups, test_msgs

def test_incompatible_version_v0_0(tmpfile):
    to_rawfile(tmpfile, version='0.0', io_version='0.0')
    assert from_rawfile(tmpfile, version='0.0')['io_version'] == '0.0'
    with pytest.raises(AssertionError):
        from_rawfile(tmpfile, version='1.0')
        pytest.fail('Should identify incompatible version')
    with pytest.raises(AssertionError):
        from_rawfile(tmpfile, version='0.1')
        pytest.fail('Should identify incompatible version')
    with pytest.raises(AssertionError):
        to_rawfile(tmpfile, io_version='0.1')
        pytest.fail('Should complain about incompatible io version')
    with pytest.raises(AssertionError):
        to_rawfile(tmpfile, io_version='1.0')
        pytest.fail('Should complain about incompatible io version')
    with pytest.raises(AssertionError):
        from_rawfile(tmpfile, io_version='0.1')
        pytest.fail('Should complain about incompatible io version')
    with pytest.raises(AssertionError):
        from_rawfile(tmpfile, io_version='1.0')
        pytest.fail('Should complain about incompatible io version')

def test_file_empty_v0_0(tmpfile):
    to_rawfile(tmpfile, version='0.0')
    assert len_rawfile(tmpfile) == 0

def test_file_full_v0_0(tmpfile, testdata):
    io_groups, msgs = testdata
    to_rawfile(tmpfile, msgs=msgs, msg_headers={'io_groups':io_groups}, version='0.0')
    assert len_rawfile(tmpfile) == len(msgs)

    rd = from_rawfile(tmpfile)
    assert len(rd['msgs']) == len(msgs)
    assert rd['msgs'][0] == msgs[0]
    assert rd['msgs'][-1] == msgs[-1]
    assert set(rd['msgs']) == set(msgs)
    assert len(rd['msg_headers']['io_groups']) == len(io_groups)
    assert rd['msg_headers']['io_groups'][0] == io_groups[0]
    assert rd['msg_headers']['io_groups'][-1] == io_groups[-1]
    assert set(rd['msg_headers']['io_groups']) == set(io_groups)

def test_file_partial_read_v0_0(tmpfile, testdata):
    io_groups, msgs = testdata
    to_rawfile(tmpfile, msgs=msgs, msg_headers={'io_groups':io_groups}, version='0.0')

    # skip reading msgs
    rd = from_rawfile(tmpfile, msg_headers_only=True)
    assert rd['msgs'] is None
    assert len(rd['msg_headers']['io_groups']) == 3

    # read from end
    rd = from_rawfile(tmpfile, start=-1)
    assert len(rd['msgs']) == 1
    assert rd['msgs'][0] == msgs[-1]
    assert len(rd['msg_headers']['io_groups']) == 1
    assert rd['msg_headers']['io_groups'][0] == io_groups[-1]

    # read from middle
    rd = from_rawfile(tmpfile, start=1, end=3)
    assert len(rd['msgs']) == 2
    assert rd['msgs'][0] == msgs[1]
    assert len(rd['msg_headers']['io_groups']) == 2
    assert rd['msg_headers']['io_groups'][0] == io_groups[1]

    # read with mask
    rd = from_rawfile(tmpfile, mask=np.array([0,1,0]).astype(bool))
    assert len(rd['msgs']) == 1
    assert rd['msgs'][0] == msgs[1]
    assert len(rd['msg_headers']['io_groups']) == 1
    assert rd['msg_headers']['io_groups'][0] == io_groups[1]

def test_file_append_v0_0(tmpfile, testdata):
    io_groups, msgs = testdata
    to_rawfile(tmpfile, msgs=msgs, msg_headers={'io_groups':io_groups}, version='0.0')
    to_rawfile(tmpfile, msgs=msgs[0:1], msg_headers={'io_groups':io_groups[0:1]}, version='0.0')
    assert len_rawfile(tmpfile) == len(msgs)+1

    rd = from_rawfile(tmpfile)
    assert len(rd['msgs']) == len(msgs)+1
    assert rd['msgs'][0] == msgs[0]
    assert rd['msgs'][-1] == msgs[0]
    assert set(rd['msgs']) == set(msgs)
    assert len(rd['msg_headers']['io_groups']) == len(io_groups)+1
    assert rd['msg_headers']['io_groups'][0] == io_groups[0]
    assert rd['msg_headers']['io_groups'][-1] == io_groups[0]
    assert set(rd['msg_headers']['io_groups']) == set(io_groups)


